package kvpaxos

import "net"
import "fmt"
import "net/rpc"
import "log"
import "paxos"
import "sync"
import "os"
import "syscall"
import "encoding/gob"
import "math/rand"
import "strconv"
import "time"

const Debug = false

func DPrintf(format string, a ...interface{}) (n int, err error) {
  if Debug {
    log.Printf(format, a...)
  }
  return
}

type Op struct {
  Type string
  Key string
  Value string
  OldValue string
  UID string
  ReqNum int
}

type Req struct {
  num int
  value string
}

type KVPaxos struct {
  mu sync.Mutex
  l net.Listener
  me int
  dead bool // for testing
  unreliable bool // for testing
  px *paxos.Paxos
  table map[string]string
  requests map[string]Req
  pxMaxSeq int
}

// try once to get/put the value in the database.  if you don't get
// agreement on your first try return an error to the client and rely
// on the client to retry.  if you need to suggest a higher sequence
// number than you have seen, update your table first.
func (kv *KVPaxos) Get(args *GetArgs, reply *GetReply) error {

  if Debug {
    fmt.Println(kv.me, " trying to get ", args.Key, " ", args.UID)
  }

  kv.mu.Lock()
  defer kv.mu.Unlock()

  kv.update()
  if kv.dead {
    reply.Err = "Did not commit"
    return nil
  }


  if Debug {
    fmt.Println(kv.me, " updated up to seq ", kv.pxMaxSeq)
  }

  uid := args.UID
  reqNum := args.ReqNum
  lastReq,seenClient := kv.requests[uid]
  if seenClient && lastReq.num > reqNum  {
      reply.Err = "Old request"
  } else if seenClient && lastReq.num == reqNum {
    if Debug {
      fmt.Println(kv.me, " already seen this operation")
    }
    reply.Value = lastReq.value
  } else {

    key := args.Key
    value := kv.table[key]

    seq := kv.pxMaxSeq + 1
    kv.pxMaxSeq = seq
    op := Op{"Get", key, value, "", uid, reqNum}

    if Debug {
      //fmt.Println(kv.me, " trying to start ", seq, " using ", op)
      fmt.Println(kv.me, " trying to start ", seq)
    }

    kv.px.Start(seq, op)
    kv.waitForStatus(seq)
    inst := kv.performOp(seq)
    kv.px.Done(seq)

    success := inst == op

    if Debug {
      fmt.Println(kv.me, " success")
    }
    if success {
      reply.Value = value
    } else {
      reply.Err = "Did not commit"
    }
  }

  return nil
}

func (kv *KVPaxos) Put(args *PutArgs, reply *PutReply) error {

  if Debug {
    //fmt.Println(kv.me, " trying to put ", args.Key, " ", args.Value, " ", args.UID)
    fmt.Println(kv.me, " trying to put ", args.Key, args.UID)
  }

  kv.mu.Lock()
  defer kv.mu.Unlock()

  kv.update()
  if kv.dead {
    reply.Err = "Did not commit"
    return nil
  }

  if Debug {
    fmt.Println(kv.me, " updated up to seq ", kv.pxMaxSeq)
  }

  uid := args.UID
  reqNum := args.ReqNum
  lastReq,seenClient := kv.requests[uid]
  if seenClient && lastReq.num > reqNum  {
      reply.Err = "Old request"
  } else if seenClient && lastReq.num == reqNum {
    if Debug {
      fmt.Println(kv.me, " already seen this operation")
    }
    reply.PreviousValue = lastReq.value
  } else {

    key := args.Key
    oldValue := kv.table[key]
    value := args.Value
    if args.DoHash {
      value = oldValue + value
      value = strconv.Itoa(int(hash(value)))
    }

    seq := kv.pxMaxSeq + 1
    kv.pxMaxSeq = seq
    op := Op{"Put", key, value, oldValue, uid, reqNum}

    if Debug {
      //fmt.Println(kv.me, " trying to start ", seq, " using ", op)
      fmt.Println(kv.me, " trying to start ", seq)
    }

    kv.px.Start(seq, op)
    kv.waitForStatus(seq)
    inst := kv.performOp(seq)
    kv.px.Done(seq)

    success := inst == op

    if success {

      if Debug {
        fmt.Println(kv.me, " success")
      }

      reply.PreviousValue = oldValue
    } else {
      reply.Err = "Did not commit"
    }
  }

  return nil
}

func (kv *KVPaxos) update() {
  pxMax := kv.px.Max()
  if pxMax > kv.pxMaxSeq {
    num := pxMax - kv.pxMaxSeq
    doneChannel := make(chan bool, num)
    for seq := kv.pxMaxSeq + 1; seq <= pxMax; seq++ {
      go func(seq int) {
        decided,_ := kv.px.Status(seq)
        if !decided {
          kv.px.Start(seq, paxos.NoOp)
          kv.waitForStatus(seq)
        }
        doneChannel <- true
      }(seq)
    }
    done := 0
    for <-doneChannel {
      done++
      if done == num {
        close(doneChannel)
        break
      }
    }
    if kv.dead {
      return
    }
    for seq := kv.pxMaxSeq + 1; seq <= pxMax; seq++ {
      kv.performOp(seq)
    }
    kv.pxMaxSeq = pxMax
    kv.px.Done(pxMax)
  }
}

func (kv *KVPaxos) performOp(seq int) Op {
  _,val := kv.px.Status(seq)
  if val == nil {
    return Op{}
  }
  var inst Op
  inst = val.(Op)
  switch inst.Type {
	default:
		panic("No such operation!")
  case "NoOp":
  case "Put":
    kv.table[inst.Key] = inst.Value
    kv.requests[inst.UID] = Req{inst.ReqNum, inst.OldValue}
  case "Get":
    kv.requests[inst.UID] = Req{inst.ReqNum, inst.Value}
  }
  return inst
}

func (kv *KVPaxos) waitForStatus(seq int) {
  to := 10 * time.Millisecond
  for !kv.dead {
    decided,_ := kv.px.Status(seq)
    if decided {
      return
    }

    if Debug {
      fmt.Println(kv.me, " haven't decided ", seq, " waiting for ", to)
    }

    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
}

// tell the server to shut itself down.
// please do not change this function.
func (kv *KVPaxos) kill() {
  DPrintf("Kill(%d): die\n", kv.me)
  kv.dead = true
  kv.l.Close()
  kv.px.Kill()
}

//
// servers[] contains the ports of the set of
// servers that will cooperate via Paxos to
// form the fault-tolerant key/value service.
// me is the index of the current server in servers[].
//
func StartServer(servers []string, me int) *KVPaxos {
  // call gob.Register on structures you want
  // Go's RPC library to marshall/unmarshall.
  gob.Register(Op{})

  kv := new(KVPaxos)
  kv.me = me

  kv.mu = sync.Mutex{}
  kv.table = make(map[string]string)
  kv.requests = make(map[string]Req)
  kv.pxMaxSeq = -1

  rpcs := rpc.NewServer()
  rpcs.Register(kv)

  kv.px = paxos.Make(servers, me, rpcs)

  os.Remove(servers[me])
  l, e := net.Listen("unix", servers[me]);
  if e != nil {
    log.Fatal("listen error: ", e);
  }
  kv.l = l


  // please do not change any of the following code,
  // or do anything to subvert it.

  go func() {
    for kv.dead == false {
      conn, err := kv.l.Accept()
      if err == nil && kv.dead == false {
        if kv.unreliable && (rand.Int63() % 1000) < 100 {
          // discard the request.
          conn.Close()
        } else if kv.unreliable && (rand.Int63() % 1000) < 200 {
          // process the request but force discard of reply.
          c1 := conn.(*net.UnixConn)
          f, _ := c1.File()
          err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
          if err != nil {
            fmt.Printf("shutdown: %v\n", err)
          }
          go rpcs.ServeConn(conn)
        } else {
          go rpcs.ServeConn(conn)
        }
      } else if err == nil {
        conn.Close()
      }
      if err != nil && kv.dead == false {
        fmt.Printf("KVPaxos(%v) accept: %v\n", me, err.Error())
        kv.kill()
      }
    }
  }()

  return kv
}
