package main

import "encoding/csv"
import "flag"
import "fmt"
import "log"
import "os"
import "strconv"

var p = flag.Int("p", 0, "persistence?")
var m = flag.Int("m", 0, "mencius?")
var c = flag.Int("c", 0, "caching?")

var throughputs = make(map[int64]map[int]*ServerThroughput)
var latencies = make(map[string]*ClientLatency)

func main() {
  flag.Parse()
  fmt.Printf("Processing throughput and latency for...\n")
  fmt.Printf("Persistence? %v\n", *p)
  fmt.Printf("Mencius? %v\n", *m)
  fmt.Printf("Caching? %v\n", *c)

  suffix := fmt.Sprintf("-p%v-m%v-c%v", *p, *m, *c)
  processThroughput(suffix)
  processLatency(suffix)
}

func processThroughput(suffix string) {
  f, err := os.Open("throughput" + suffix + ".csv")
  defer f.Close()
  if err != nil {
    log.Fatal(err)
  }

  reader := csv.NewReader(f)
  records, err := reader.ReadAll()
  if err != nil {
    log.Fatal(err)
  }
  
  for _, record := range records {
    // kv.gid, kv.me, seq, inst.Type, start, end, elapsed
    gid, _ := strconv.ParseInt(record[0], 10, 64)
    me64, _ := strconv.ParseInt(record[1], 10, 0)
    me := int(me64)

    _, ok := throughputs[gid]
    if !ok {
      throughputs[gid] = make(map[int]*ServerThroughput)
    }
    _, ok = throughputs[gid][me]
    if !ok {
      st := &ServerThroughput{}
      st.gid = gid
      st.me = me
      st.bins = make([]int, 0)
      throughputs[gid][me] = st
    }
  }

  for gid, _ := range throughputs {
    for me, _ := range throughputs[gid] {
      bins := throughputs[gid][me].bins

      for _, record := range records {
        gid2, _ := strconv.ParseInt(record[0], 10, 64)
        me64, _ := strconv.ParseInt(record[1], 10, 0)
        me2 := int(me64)
        if gid2 != gid || me2 != me {
          continue
        }
        
        reqtype := record[3]
        if reqtype == "Reconfigure" {
          continue
        }
        
        end, err := strconv.ParseInt(record[5], 10, 64)
        if err != nil {
          log.Fatal(err)
        }

        bucket := int(end / 1e9)
        for bucket > len(bins) - 1 {
          bins = append(bins, 0)
        }
        bins[bucket]++
      }

      throughputs[gid][me].bins = bins
    }
  }

  for gid, _ := range throughputs {
    for me, _ := range throughputs[gid] {
      st := throughputs[gid][me]
      
      if len(st.bins) < 3 {
        st.t = 0
        continue
      }
      
      var sum_t int
      var count int
      for i := 1; i < len(st.bins)-1; i++ {
        if st.bins[i] > 0 {
          sum_t += st.bins[i]
          count++
        }
      }
      st.t = float64(sum_t) / float64(count)
    }
  }
  
  fmt.Println()
  fmt.Printf("gid[me]     throughput\n")
  var sum float64
  for gid, _ := range throughputs {
    for me, _ := range throughputs[gid] {
      st := throughputs[gid][me]
      fmt.Printf("%v[%v]        %.2f\n", st.gid, st.me, st.t)
      sum += st.t
    }
  }
  fmt.Printf("average       %.2f\n", sum / float64(len(throughputs[100])))
  
  maxtime := -1
  for gid, _ := range throughputs {
    for me, _ := range throughputs[gid] {
      st := throughputs[gid][me]
      if len(st.bins) - 1 > maxtime {
        maxtime = len(st.bins) - 1
      }
    }
  }
  for gid, _ := range throughputs {
    for me, _ := range throughputs[gid] {
      st := throughputs[gid][me]
      
      serversuffix := fmt.Sprintf("-[%v][%v]", gid, me)
      f, err := os.Create("latency" + suffix + serversuffix + ".csv")
      defer f.Close()
      if err != nil {
        log.Fatal(err)
      }
      l := log.New(f, "", 0)
      
      for i := 0; i < maxtime; i++ {
        if i >= len(st.bins) {
          l.Printf("%v,%v\n", i, 0)
        } else {
          l.Printf("%v,%v\n", i, st.bins[i])
        }
      }
    }
  }
}

type ServerThroughput struct {
  gid int64
  me int
  bins []int
  t float64
}

func processLatency(suffix string) {
  f, err := os.Open("latency" + suffix + ".csv")
  defer f.Close()
  if err != nil {
    log.Fatal(err)
  }

  reader := csv.NewReader(f)
  records, err := reader.ReadAll()
  if err != nil {
    log.Fatal(err)
  }

  for _, record := range records {
    // ck.uid, ck.reqNum, reqtype, start, end, elapsed
    uid := record[0]
    reqtype := record[2]
    elapsed, err := strconv.ParseInt(record[5], 10, 64)
    if err != nil {
      log.Fatal(err)
    }

    stat, ok := latencies[uid]
    if !ok {
      stat = &ClientLatency{}
      stat.uid = uid
      latencies[uid] = stat
    }
    switch reqtype {
    case "Get":
      stat.lga = append(stat.lga, elapsed)
    case "Put":
      stat.lpa = append(stat.lpa, elapsed)
    }
  }

  for _, stat := range latencies {
    var sum_lg int64
    var sum_lp int64
    for _, l := range stat.lga {
      sum_lg += l
    }
    for _, l := range stat.lpa {
      sum_lp += l
    }

    stat.ng = int64(len(stat.lga))
    stat.np = int64(len(stat.lpa))
    stat.n = stat.ng + stat.np
    stat.lg = sum_lg / stat.ng
    stat.lp = sum_lp / stat.np
    stat.l = (sum_lg + sum_lp) / stat.n
  }

  var sum_lg, sum_lp, sum_l int64
  var ng, np, n int64
  for _, stat := range latencies {
    sum_lg += stat.lg * stat.ng
    sum_lp += stat.lp * stat.np
    sum_l += stat.l * stat.n
    ng += stat.ng
    np += stat.np
    n += stat.ng + stat.np
  }
  lg := sum_lg / ng
  lp := sum_lp / np
  l := sum_l / n

  fmt.Println()
  fmt.Printf("client     get latency   put latency   overall latency\n")
  for _, stat := range latencies {
    fmt.Printf("%v...    %.5f       %.5f       %.5f\n",
      stat.uid[0:4],
      float64(stat.lg) / float64(1e9),
      float64(stat.lp) / float64(1e9),
      float64(stat.l) / float64(1e9))
  }
  fmt.Printf("%v    %.5f       %.5f       %.5f\n",
    "average",
    float64(lg) / float64(1e9),
    float64(lp) / float64(1e9),
    float64(l) / float64(1e9))
}

type ClientLatency struct {
  uid string
  lga []int64
  lpa []int64
  lg int64
  lp int64
  l int64
  ng int64
  np int64
  n int64
}
