package main

import "encoding/gob"
import "fmt"
import "io/ioutil"
import "os"
import "os/exec"
import "shardmaster"
import "strconv"

func main() {
  fmt.Println("Test nil slice...")
  testNilSlice()

  fmt.Println("Test type conversion...")
  testTypeConversion()

  fmt.Println("Test read write...")
  testReadWrite()

  fmt.Println("Test move dir...")
  testMoveDir()

  fmt.Println("Test string format...")
  testStringFormat()

  fmt.Println("Test gob encoding...")
  testGobEncoding()

  fmt.Println("Test simple gob...")
  testSimpleGob()
}

func testNilSlice() {
  var a []int
  a = nil
  a = append(a, 2)
  fmt.Println(a)
}

func testTypeConversion() {
  var v interface{}

  type Op struct {
    x int
    y int
  }

  var o Op
  o.x = 1
  o.y = 2

  v = o

  fmt.Println(v)

  // This code is illegal.
  // var b []byte
  // b = v.([]byte)
  // fmt.Println(b)
}

func testReadWrite() {
  var err error
  var f *os.File
  var fi os.FileInfo
  var data []byte
  var n int

  // Test ioutil's WriteFile and ReadFile functions.
  for i := 0; i < 10; i++ {
    data = append(data, byte('a'+i))
  }
  fmt.Println(data)
  ioutil.WriteFile("/tmp/test", data, 0666)
  data, err = ioutil.ReadFile("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Println(data)

  // Test conversion between string and []byte.
  var s string
  var b []byte
  for i := 0; i < 10; i++ {
    b = append(b, byte('0'+i))
  }
  s = string(b)
  b = []byte(s)
  fmt.Println(s)
  fmt.Println(b)

  // Test os's file reading and writing functions.
  f, err = os.Open("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fi, err = f.Stat()
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  data = make([]byte, fi.Size())
  n, err = f.Read(data)
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Printf("read %v bytes, %v\n", n, data[:n])
  f.Close()

  f, err = os.Create("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  n, err = f.Write(b)
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Printf("wrote %v bytes, %v\n", n, b)
  n, err = f.WriteString(s)
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Printf("wrote %v bytes, %v\n", n, s)
  f.Sync()
  f.Close()

  f, err = os.Open("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fi, err = f.Stat()
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  data = make([]byte, fi.Size())
  n, err = f.Read(data)
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Printf("read %v bytes, %v\n", n, data[:n])
  f.Close()

  // Test removing files and errors.
  err = os.Remove("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  fmt.Printf("removed file /tmp/test\n")
  err = os.Remove("/tmp/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }

  // Test make file without existing directory.
  f, err = os.Create("/tmp/6.824-ksw/scratch/test")
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }
  n, err = f.Write(b)
  if err != nil {
    fmt.Printf("err: %v\n", err)
  }

  err = os.Chdir("/tmp/6.824-ksw/scratch")
  if err != nil {
    fmt.Printf("err: %v\n", err)
    fmt.Printf("because didn't exist? %v\n", os.IsNotExist(err))
  }

}

func testMoveDir() {
  os.RemoveAll("/tmp/6.824-ksw/scratch")
  os.RemoveAll("/tmp/6.824-ksw/scratch2")
  os.Mkdir("/tmp/6.824-ksw/scratch", 0777)
  os.Mkdir("/tmp/6.824-ksw/scratch2", 0777)
  os.Chdir("/tmp/6.824-ksw/scratch")
  os.Mkdir("test-dir", 0777)

  os.Chdir("test-dir")
  f, _ := os.Create("test-file")
  f.WriteString("test-string\n")
  f.Close()
  os.Mkdir("inner-dir", 0777)

  //os.Rename("/tmp/6.824-ksw/scratch/test-dir", "/tmp/6.824-ksw/scratch2/test-dir")
  os.Chdir("..")
  os.Rename("test-dir", "../scratch2/test-dir")

  dir, _ := os.Open("/afs/athena.mit.edu/user/w/a/wangaj/6.824-project/src")
  names, _ := dir.Readdirnames(0)
  for i, n := range names {
    fmt.Printf("name %v is %v\n", i, n)
  }
  dir.Close()
  fmt.Printf("printed %v names\n", len(names))

  cmd := exec.Command("cp", "-r",
    "/tmp/6.824-ksw/scratch2",
    "/tmp/6.824-ksw/scratch/scratch2")
  out, err := cmd.Output()
  fmt.Println(out)
  fmt.Println(err)
}

func testStringFormat() {
  var cf shardmaster.Config
  cf.Num = 1
  for i, _ := range cf.Shards {
    cf.Shards[i] = 10 * int64(i)
  }
  cf.Groups = make(map[int64][]string)
  cf.Groups[0] = []string{"00", "01", "02"}
  cf.Groups[1] = []string{"10", "11", "12"}
  cf.Groups[2] = []string{"20", "21", "22"}
  fmt.Println(cf.Num)
  fmt.Println(cf.Shards)
  fmt.Println(cf.Groups)
  fmt.Println(cf)
}

func testGobEncoding() {
  type A struct {
    A int
    B string
    C map[int]string
  }
  gob.Register(A{})

  a := &A{}
  a.A = 2
  a.B = "hello"
  a.C = make(map[int]string)
  a.C[3] = "goodbye"
  for i := 4; i < 10; i++ {
    a.C[i] = strconv.Itoa(i) + strconv.Itoa(i)
  }

  b := &A{}
  b.A = 12
  b.B = "gob"
  b.C = make(map[int]string)
  b.C[13] = "encoding"
  for i := 4; i < 10; i++ {
    b.C[i] = strconv.Itoa(i) + strconv.Itoa(i)
  }

  os.Mkdir("/tmp/6.824-ksw/scratch", 0777)
  os.Chdir("/tmp/6.824-ksw/scratch")
  f, _ := os.Create("output")

  enc := gob.NewEncoder(f)
  err := enc.Encode(a)
  if err != nil {
    fmt.Println(err)
  }
  err = enc.Encode(b)
  if err != nil {
    fmt.Println(err)
  }
  f.Sync()
  f.Close()

  f, _ = os.Open("output")
  ra := &A{}
  rb := &A{}
  dec := gob.NewDecoder(f)

  readValue(dec, ra)
  fmt.Println(ra)
  readValue(dec, rb)
  fmt.Println(rb)

  f.Close()
}

func readValue(dec *gob.Decoder, value interface{}) {
  err := dec.Decode(value)
  if err != nil {
    fmt.Printf("readValue() error: %v\n", err)
  }
}

func testSimpleGob() {
  os.Mkdir("/tmp/6.824-ksw/scratch", 0777)
  os.Chdir("/tmp/6.824-ksw/scratch")
  f, _ := os.Create("output-simple")

  var a int
  var b int
  a = 2
  b = 3

  enc := gob.NewEncoder(f)
  err := enc.Encode(a)
  if err != nil {
    fmt.Println(err)
  }
  err = enc.Encode(b)
  if err != nil {
    fmt.Println(err)
  }
  f.Sync()
  f.Close()

  f, _ = os.Open("output-simple")
  var ra int
  var rb int

  dec := gob.NewDecoder(f)
  err = dec.Decode(&ra)
  fmt.Println(ra)
  err = dec.Decode(&rb)
  fmt.Println(rb)

  f.Close()
}
