package paxos

//
// Paxos library, to be included in an application.
// Multiple applications will run, each including
// a Paxos peer.
//
// Manages a sequence of agreed-on values.
// The set of peers is fixed.
// Copes with network failures (partition, msg loss, &c).
// Does not store anything persistently, so cannot handle crash+restart.
//
// The application interface:
//
// px = paxos.Make(peers []string, me string)
// px.Start(seq int, v interface{}) -- start agreement on new instance
// px.Status(seq int) (Decided bool, v interface{}) -- get info about an instance
// px.Done(seq int) -- ok to forget all instances <= seq
// px.Max() int -- highest instance seq known, or -1
// px.Min() int -- instances before this seq have been forgotten
//

import "encoding/gob"
import "fmt"
import "log"
import "math/rand"
import "net"
import "net/rpc"
import "os"
import "sort"
import "strconv"
import "strings"
import "sync"
import "syscall"
import "time"
import "util"
import "os/exec"


var debug bool = false //true
var debugDisk bool = false //true
var useNew = true


var NoOp = util.Op{Type: NoOpType}

type Paxos struct {
  dirpath string
  mu sync.Mutex
  muDisk sync.Mutex
  muHD sync.Mutex
  l net.Listener
  dead bool
  unreliable bool
  rpcCount int
  peers []string
  me int // index into peers[]
  instances map[int]*Instance
  seqMax int
  seqMins []int
	failures map[int64]int
  cleaned int

  hasDisk bool
  hasDiskSet bool

  menciusPtr int // next owned mencius instance TODO init to me
  // skipList []int // instances this node is skipping
  skipList map[int]bool
  failDetection map[int]int

	Debug bool
	Persist bool
  Mencius bool
}

type Instance struct {
  PropMax int
  AcceptMax int
  AcceptValue interface{}
  DecideValue interface{}
  Decided bool
  Proposing bool
}

func (px *Paxos) GetInstance(seq int) *Instance {
  //px.MinHelper()
  //fmt.Println(px.peers[px.me], " Get Instance1")
  if seq < px.MinHelper() {
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "trying to get:", seq, "but min is", px.MinHelper())
    }
    //fmt.Println(px.peers[px.me], " Get Instance2")
    return nil
  } else {
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "trying to get:", seq)
    }
    //fmt.Println(px.peers[px.me], " Get Instance3")
    if seq > px.seqMax {
      px.seqMax = seq
    }
    //fmt.Println(px.peers[px.me], " Get Instance4")
    inst, ok := px.instances[seq]
    if !ok {
      if px.Mencius {
        px.instances[seq] = &Instance{SUGGEST_SEQ, -1, nil, nil, false, false} // &Instance{-1, -1, nil, nil, false, false}
        inst = px.instances[seq]
      } else {
        px.instances[seq] = &Instance{-1, -1, nil, nil, false, false}
        inst = px.instances[seq]
      }
    }
          //fmt.Println(px.peers[px.me], "trying to get:", seq, inst)

    return inst
  }
}

func (px *Paxos) UpdateSeqMins(you int, yourFailures map[int64]int, yourVersion []int) {
	// var newFailSeq int
	// var newFail bool
	// for k,v := range yourFailures {
	// 	if _,ok := px.myFailures[k]; !ok {			
	// 		if newFail {
	// 			if v < newFailSeq {
	// 				newFailSeq = v
	// 			}
	// 		} else {
	// 			newFail = true
	// 			newFailSeq = v
	// 		}
	// 		px.myFailures[k] = v
	// 	}
	// }
	// if newFail {
	// 	newMin := math.Min(newFailSeq, px.Min() - 1)
	// 	for i,_ := range px.seqMins {
	// 		px.seqMins[i] = newMin
	// 	}
	// } else {
	// 	if len(px.myFailures) == 0 {

//	for i, seq := range yourVersion {
//		if seq > px.seqMins[i] {
//			px.seqMins[i] = seq
//		}
//	}
	
	// 	} else {
	if yourVersion[you] > px.seqMins[you] {
		px.seqMins[you] = yourVersion[you]
	}
	// 	}
	// }
	px.UpdateMinState()
}

func (px *Paxos) UpdateMinState() {
  min := px.MinHelper()
  for seq := px.cleaned + 1; seq < min; seq++ {
    if debug || px.Debug {
      fmt.Println("deleting seq", seq)
    }
    delete(px.instances, seq)
  }
  px.cleaned = min - 1

	if px.Persist {
		px.muDisk.Lock()
		px.deleteOldDisk()
		px.saveDoneState()
		px.muDisk.Unlock()
	}
}

func (px *Paxos) Prepare(args *PrepareArgs, reply *PrepareReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  if px.dead {
    return nil
  }

  px.UpdateSeqMins(args.Me, args.Failures, args.SeqMins)
  reply.SeqMins = px.seqMins
	reply.Failures = px.failures
	reply.Me = px.me

  if px.Mencius {
    //px.failDetection[args.Me] = args.MenciusPtr
    px.processMenciusPtrs(args.MenciusPtr)
  }

  seq := args.SeqNum
  n := args.PropNum
  inst := px.GetInstance(seq)
  if inst == nil {
    return nil
  }

  if debug || px.Debug {
    fmt.Println(px.peers[px.me], "prepare:", "seq", seq, "prop", n,
      "pm", inst.PropMax, "am", inst.AcceptMax)
  }

  if n > inst.PropMax {
    inst.PropMax = n
    reply.OK = true
    reply.AcceptMax = inst.AcceptMax
    reply.AcceptValue = inst.AcceptValue
    //fmt.Println(px.peers[px.me], "prepare reply ", reply)
  } else {
    reply.PropMax = inst.PropMax
  }

  if px.Mencius {
    reply.MenciusPtr = px.failDetection //px.menciusPtr
  } 

	if px.Persist {
		px.muDisk.Lock()
		px.saveAcceptorState(args.SeqNum)
		px.muDisk.Unlock()
	}

  return nil
}

func (px *Paxos) Accept(args *AcceptArgs, reply *AcceptReply) error {
  px.mu.Lock()
	defer px.mu.Unlock()
  if px.dead {
    return nil
  }

  px.UpdateSeqMins(args.Me, args.Failures, args.SeqMins)
  reply.SeqMins = px.seqMins
	reply.Failures = px.failures
	reply.Me = px.me

  if px.Mencius {
    //px.failDetection[args.Me] = args.MenciusPtr
    px.processMenciusPtrs(args.MenciusPtr)
  }
  seq := args.SeqNum
  n := args.PropNum
  val := args.Val
  inst := px.GetInstance(seq)
  if inst == nil {
    return nil
  }

  
  if px.Mencius {
    if args.Suggest {
      px.insertMySkips(seq)
    }
    px.processSkipList(args.SkipList)
  }

  if debug || px.Debug {
    fmt.Println(px.peers[px.me], "accept: ", "seq", seq, "prop", n,
      "pm", inst.PropMax, "am", inst.AcceptMax, inst)
  }


  if n >= inst.PropMax  {
    inst.PropMax = n
    inst.AcceptMax = n
    inst.AcceptValue = val
    reply.OK = true
    reply.PropNum = n

    // TODO reply.menciusPts
  } else {
    reply.PropMax = inst.PropMax
  }

  if px.Mencius {
    reply.SkipList = px.skipList
    //reply.MenciusPtr = px.menciusPtr
    reply.MenciusPtr = px.failDetection
  }  

	if px.Persist {
    //fmt.Println("waiting for muDisk")
		px.muDisk.Lock()
		px.saveAcceptorState(args.SeqNum)
		px.muDisk.Unlock()
    //fmt.Println("done waiting for muDisk")
	}

  return nil
}

func (px *Paxos) Decide(args *DecideArgs, reply *DecideReply) error {
  px.mu.Lock()
	defer	px.mu.Unlock()
  if px.dead {
    return nil
  }

  px.UpdateSeqMins(args.Me, args.Failures, args.SeqMins)
  reply.SeqMins = px.seqMins
	reply.Failures = px.failures
	reply.Me = px.me

  if px.Mencius {
    //px.failDetection[args.Me] = args.MenciusPtr
    px.processMenciusPtrs(args.MenciusPtr)
    px.processSkipList(args.SkipList)
  }

  seq := args.SeqNum
  n := args.PropNum
  val := args.Val
  inst := px.GetInstance(seq)
  if inst == nil {
    return nil
  }

  if px.Mencius {
    if args.Suggest {
      px.insertMySkips(seq)
    }
    px.processSkipList(args.SkipList)
  }
  

  if debug || px.Debug {
    fmt.Println(px.peers[px.me], "decide: ", "seq", seq, "prop", n,
      "pm", inst.PropMax, "am", inst.AcceptMax, "val", val)

  }

  if n >= inst.AcceptMax {
    inst.AcceptMax = n
    inst.AcceptValue = val
  }
  if n > inst.PropMax {
    inst.PropMax = n
  }

  inst.DecideValue = val
  inst.Decided = true

  if px.Mencius {
    inst.PropMax += 1
    if px.Persist {
      px.muDisk.Lock()
      px.saveAcceptorState(args.SeqNum)
      //fmt.Println(px.peers[px.me], "saved decide: ", "seq", seq, "inst", inst)
      px.muDisk.Unlock()
    }
  }

  if px.Mencius {
    reply.SkipList = px.skipList
    //reply.MenciusPtr = px.menciusPtr
    reply.MenciusPtr = px.failDetection
  }
  return nil
}

//
// call() sends an RPC to the rpcname handler on server srv
// with arguments args, waits for the reply, and leaves the
// reply in reply. the reply argument should be a pointer
// to a reply structure.
//
// the return value is true if the server responded, and false
// if call() was not able to contact the server. in particular,
// the replys contents are only valid if call() returned true.
//
// you should assume that call() will time out and return an
// error after a while if it does not get a reply from the server.
//
// please use call() to send all RPCs, in client.go and server.go.
// please do not change this function.
//
func call(srv string, name string, args interface{}, reply interface{}) bool {
  c, err := rpc.Dial("unix", srv)
  if err != nil {
    err1 := err.(*net.OpError)
    if err1.Err != syscall.ENOENT && err1.Err != syscall.ECONNREFUSED {
      fmt.Printf("paxos Dial() failed: %v\n", err1)
    }
    return false
  }
  defer c.Close()

  err = c.Call(name, args, reply)
  if err == nil {
    return true
  }

  fmt.Println(err)
  return false
}


//
// the application wants paxos to start agreement on
// instance seq, with proposed value v.
// Start() returns right away; the application will
// call Status() to find out if/when agreement
// is reached.
//
func (px *Paxos) Start(seq int, v interface{}) {

  go func() {


    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start1:", "seq", seq, "val", v)
    }
    // px.mu.Unlock()
    px.mu.Lock()

    inst := px.GetInstance(seq)

    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start2:", "seq", seq)
    }
    px.mu.Unlock()


    if inst == nil {
      if debug || px.Debug {
        fmt.Println(px.peers[px.me], "start:", "seq", seq, "inst", nil)
      }
      return
    }
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start3:", "seq", seq)
    }

    if px.Mencius {
      // if this is an owned instance that has not yet started
      //    then send a suggest
      if px.ownInstance(seq) && inst.Proposing == false {
        n_suggest := seq
        if seq == px.menciusPtr { // check shouldn't be necessary
          px.mu.Lock()
          n_suggest = px.menciusPtr
          px.menciusPtr += len(px.peers)
          px.muDisk.Lock()
          px.saveMenciusState()
          px.muDisk.Unlock()
          px.mu.Unlock()
        }

        if px.AttemptToSuggest(seq, n_suggest, inst.PropMax, v) {
          return
        }
      }
    }

    px.mu.Lock()
    inst.Proposing = true
    px.mu.Unlock()

    for !inst.Decided && !px.dead {

      n := GetNextNum(inst.PropMax, px.me, len(px.peers))

      if debug || px.Debug {
        fmt.Println(px.peers[px.me], "start4:", "seq", seq, "prop", n)
      }

      prepared, PropMax, val := px.AttemptToAccept(seq, n, inst.PropMax, v)
      px.mu.Lock()
      if PropMax > inst.PropMax {
        inst.PropMax = PropMax
      }
      px.mu.Unlock()

      if debug || px.Debug {
        fmt.Println(px.peers[px.me], "start attempt to accept:", "seq", seq, "prop", n, "prepared", prepared, "val", val)
      }

      if prepared && !px.dead {

        accepted, PropMax := px.AttemptToDecide(seq, n, PropMax, val, false)

        px.mu.Lock()
        if PropMax > inst.PropMax {
          inst.PropMax = PropMax
        }
        px.mu.Unlock()

        if debug || px.Debug {
          fmt.Println(px.peers[px.me], "start:", "seq", seq, "prop", n, "accepted", accepted)
        }

        if accepted && !px.dead {
          px.SendDecideToAll(seq, n, val, false)
        }
      }

      px.mu.Lock()
      inst = px.GetInstance(seq)
      px.mu.Unlock()
      if inst == nil {
        return
      }
    }

		return
  }()
}

// only called from within mencius flag
func (px *Paxos) AttemptToSuggest(seq int, propNum int, PropMax int, val interface{}) (bool)  {

  px.mu.Lock()
  inst := px.GetInstance(seq)
  inst.Proposing = true
  px.mu.Unlock()

  n := SUGGEST_SEQ //GetNextNum(inst.propMax, px.me, len(px.peers))

  if debug {
   fmt.Println(px.me, "suggest:", "seq", seq, "prop", n, " Instance struct ", inst)
  }

  if !px.dead {

    accepted, PropMax := px.AttemptToDecide(seq, n, PropMax, val, true)
    px.mu.Lock()
    if PropMax > inst.PropMax {
      inst.PropMax = PropMax
    }
    px.mu.Unlock()

    if debug {
      fmt.Println(px.me, "start:", "seq", seq, "prop", n, "accepted", accepted)
    }

    if accepted && !px.dead {
      px.SendDecideToAll(seq, n, val, true)

      //fmt.Println(px.me, "suggest succeeded:", "seq", seq, "prop", n)
      return true
    }
  }

  px.mu.Lock()
  inst = px.GetInstance(seq)
  px.mu.Unlock()
  if inst == nil {
    return false // TODO ?
  }
  return false
}

func (px *Paxos) AttemptToAccept(seq int, propNum int, PropMax int, val interface{}) (bool, int, interface{}) {

  accepted := 0
  rejected := 0
  AcceptMax := -1

  args := &PrepareArgs{seq, propNum, px.seqMins, px.failDetection, px.me, px.failures}

  for i := range px.peers {
    reply := &PrepareReply{}
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start AttemptToAccept:", "seq", seq, "prop", propNum, "i", i)
    }
    if i == px.me {
      px.Prepare(args, reply)
    } else {
      //fmt.Println(px.peers[px.me], "calling :", "seq", seq, "prop", propNum, "i", i)
      call(px.peers[i], "Paxos.Prepare", args, reply)
    }
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start:", "seq", seq, "prop", propNum, "i", i)
    }
    px.mu.Lock()
    if px.dead {
			px.mu.Unlock()
      break
    }

    //px.UpdateSeqMins(reply.SeqMins)
    if px.Mencius {
      //px.failDetection[i] = reply.MenciusPtr
      px.processMenciusPtrs(reply.MenciusPtr)
    }

		if reply.OK {
			px.UpdateSeqMins(reply.Me, reply.Failures, reply.SeqMins)
		}

    px.mu.Unlock()
    if debug || px.Debug {
      fmt.Println(px.peers[px.me], "start:", "seq", seq, "prop", propNum, "i", i)
    }
    if reply.OK {
      accepted++
      //fmt.Println(px.peers[px.me], "start check for accept max in attempt to accept:", "seq", seq, "prop", propNum, "i", i, "accept max", AcceptMax, "reply", reply)
      if px.Mencius{
        if reply.AcceptMax >= AcceptMax {
          AcceptMax = reply.AcceptMax
          val = reply.AcceptValue
        }
      } else {
        if reply.AcceptMax > AcceptMax {
          AcceptMax = reply.AcceptMax
          val = reply.AcceptValue
        }
      }
    } else {
      rejected++
      if reply.PropMax > PropMax {
        PropMax = reply.PropMax
      }
    }
    if accepted > len(px.peers)/2 {
      return true, PropMax, val
    }
    if rejected >= (len(px.peers) + 1)/2 {
      return false, PropMax, val
    }
  }
  return false, -1, nil
}


func (px *Paxos) AttemptToDecide(seq int, propNum int, PropMax int, v interface{}, suggest bool) (bool, int) {

  accepted := 0
  rejected := 0

  args := &AcceptArgs{seq, propNum, v, px.seqMins, suggest, px.skipList, px.failDetection, px.me, px.failures}

  for i := range px.peers {
    reply := &AcceptReply{}
    if i == px.me {
      px.Accept(args, reply)
    } else {
      call(px.peers[i], "Paxos.Accept", args, reply)
    }

    px.mu.Lock()
    if px.Mencius {
      px.processSkipList(reply.SkipList)
      //px.failDetection[i] = reply.MenciusPtr
      px.processMenciusPtrs(reply.MenciusPtr)
    }
    if px.dead {
			px.mu.Unlock()
      break
    }

		if reply.OK {
			px.UpdateSeqMins(reply.Me, reply.Failures, reply.SeqMins)
		}

    px.mu.Unlock()
    if reply.OK {
      accepted++
    } else {
      rejected++
      if reply.PropMax > PropMax {
        PropMax = reply.PropMax
      }
    }
    if accepted > len(px.peers)/2 {
      return true, PropMax
    }
    if rejected >= (len(px.peers) + 1)/2 {
      return false, PropMax
    }
  }
  return false, -1
}


func (px *Paxos) SendDecideToAll(seq int, propNum int, val interface{}, suggest bool) {
  args := &DecideArgs{seq, propNum, val, px.seqMins, suggest, px.skipList, px.failDetection, px.me, px.failures}

  for i := range px.peers {
    reply := &DecideReply{}
    if i == px.me {
      px.Decide(args, reply)
    } else {
      call(px.peers[i], "Paxos.Decide", args, reply)
    }

    px.mu.Lock()
    if px.Mencius {
      px.processSkipList(reply.SkipList)
      //px.failDetection[i] = reply.MenciusPtr
      px.processMenciusPtrs(reply.MenciusPtr)
    }
    px.mu.Unlock()
  }
}

func (px *Paxos) MonitorDecisions() {
  if px.Mencius == false {
    for !px.dead {
      //fmt.Println("Monitor", px.me)
      px.Monitor()
      time.Sleep(1 * time.Second)
    }
  }
}

func (px *Paxos) Monitor() {
  px.mu.Lock()
  defer px.mu.Unlock()
  if px.dead {
    return
  }

  for seq := px.seqMins[px.me] + 1; seq <= px.seqMax; seq++ {
    inst := px.GetInstance(seq)
    if inst != nil && !inst.Decided && !inst.Proposing {
      px.Start(seq, NoOp)
    }
    //fmt.Println("Monitor loop", px.me)
  }
  //fmt.Println("Monitor", px.me)
}

func (px *Paxos) Print() {
  keys := make([]int, len(px.instances))
  i := 0
  for k, _ := range px.instances {
    keys[i] = k
    i++
  }
  sort.Ints( keys )
  fmt.Println(px.peers[px.me], "min", px.MinHelper(), "keys", keys)
}

// return a sequence number greater than my current max
func (px *Paxos) GetNextSlot() int {
  // next owned slot greater than max incorrect due to bulk no-ops
  // need to maintain mencius pointer
  px.mu.Lock()
  defer px.mu.Unlock()

  if px.Mencius {
	  return px.menciusPtr // len(px.peers)
  } else {
    return px.seqMax + 1
  }
}

//
// the application on this machine is done with
// all instances <= seq.
//
// see the comments for Min() for more explanation.
//
func (px *Paxos) Done(seq int) {
  px.mu.Lock()
  defer px.mu.Unlock()
  if px.dead {
    return
  }

//	fmt.Println(px.peers[px.me], "Done", seq)
  if seq > px.seqMins[px.me] {
    px.seqMins[px.me] = seq
  }
  px.UpdateMinState()
}

//
// the application wants to know the
// highest instance sequence known to
// this peer.
//
func (px *Paxos) Max() int {
  px.mu.Lock()
  defer px.mu.Unlock()
  return px.seqMax
}

//
// Min() should return one more than the minimum among z_i,
// where z_i is the highest number ever passed
// to Done() on peer i. A peers z_i is -1 if it has
// never called Done().
//
// Paxos is required to have forgotten all information
// about any instances it knows that are < Min().
// The point is to free up memory in long-running
// Paxos-based servers.
//
// Paxos peers need to exchange their highest Done()
// arguments in order to implement Min(). These
// exchanges can be piggybacked on ordinary Paxos
// agreement protocol messages, so it is OK if one
// peers Min does not reflect another Peers Done()
// until after the next instance is agreed to.
//
// The fact that Min() is defined as a minimum over
// *all* Paxos peers means that Min() cannot increase until
// all peers have been heard from. So if a peer is dead
// or unreachable, other peers Min()s will not increase
// even if all reachable peers call Done. The reason for
// this is that when the unreachable peer comes back to
// life, it will need to catch up on instances that it
// missed -- the other peers therefor cannot forget these
// instances.
//
func (px *Paxos) Min() int {
  px.mu.Lock()
  px.mu.Unlock()
  //fmt.Println(" Min called on ", px.me," with list ", px.seqMins)
  globalSeqMin/*, ok */:= px.seqMins[px.me]
  //if !ok {
  //  px.seqMins[px.me] = -1
  //}
  for _, seq := range px.seqMins {
    if seq < globalSeqMin {
      globalSeqMin = seq
    }
  }
  return globalSeqMin + 1
}

func (px *Paxos) MinHelper() int {

  //fmt.Println(" Min called on ", px.me," with list ", px.seqMins)
  globalSeqMin/*, ok */:= px.seqMins[px.me]
  //if !ok {
  //  px.seqMins[px.me] = -1
  //}
  for _, seq := range px.seqMins {
    if seq < globalSeqMin {
      globalSeqMin = seq
    }
  }
  return globalSeqMin + 1
}

//
// the application wants to know whether this
// peer thinks an instance has been Decided,
// and if so what the agreed value is. Status()
// should just inspect the local peer state;
// it should not contact other Paxos peers.
//
func (px *Paxos) Status(seq int) (bool, interface{}) {
  px.mu.Lock()
  defer px.mu.Unlock()

  if px.Mencius {
    if px.dead == false {
    	inst, ok := px.instances[seq]
      if !ok {
        return false, nil
      } else {
        return inst.Decided, inst.DecideValue
      }
    } else {
      //fmt.Println("Dead instance status in seq ", seq)
      return false, nil
    }
  } else {
    // if debug {
    if seq < px.MinHelper() {
      fmt.Println(px.peers[px.me], "Wanted", seq, "but min was", px.MinHelper())
    }
    // }
    inst, ok := px.instances[seq]
    if !ok {
      // fmt.Println(px.peers[px.me], "status", seq, px.Min(), px.seqMins)
      return false, nil
    } else {
      return inst.Decided, inst.DecideValue
    }
  }
}




//
// tell the peer to shut itself down.
// for testing.
// please do not change this function.
//
func (px *Paxos) Kill(keepdisk ...bool) {
  px.mu.Lock()
  defer px.mu.Unlock()

  px.dead = true
  if px.l != nil {
    px.l.Close()
  }

	for k,_ := range px.instances {
		delete(px.instances, k)
	}

	if px.Persist {
		if len(keepdisk) > 0 && keepdisk[0] == true {
			// do nothing to the disk
		} else {
			px.muDisk.Lock()
			px.loseDisk()
			px.muDisk.Unlock()
		}
	}

	if debug || px.Debug {
		fmt.Printf("killed Paxos[%v]\n", px.me)
	}
}

func (px *Paxos) GetSeqMins() []int {
	return px.seqMins
}

//
// the application wants to create a paxos peer.
// the ports of all the paxos peers (including this one)
// are in peers[]. this servers port is peers[me].
//
func Make(mencius bool, persist bool, peers []string, me int, rpcs *rpc.Server) *Paxos {
  gob.Register(util.Op{})
  gob.Register(Instance{})

  px := &Paxos{}
  px.peers = peers
  px.me = me
  px.mu = sync.Mutex{}
  px.muDisk = sync.Mutex{}
	px.Persist = persist
  px.Mencius = mencius

  if px.Mencius {
  	if px.Persist {
  		// Restore persistent state if it exists.
      px.mu.Lock()
      go func() {
          //fmt.Println(px.me, " attempting to restore")
          px.restore(peers, me)
          // use fetched mencius state correctly
          if px.menciusPtr % len(peers) != px.me {
            //fmt.Println(px.me, " need to reset menciusptr")
            px.resetMenciusPtr()
            //fmt.Println("need to reset menciusptr ... past reset")
            px.muDisk.Lock()
            px.saveMenciusState()
            px.muDisk.Unlock()
            //fmt.Println(px.me, " need to reset menciusptr ... complete")
          }
          //fmt.Println(px.me, " used restore")
        px.mu.Unlock()
      }()
  	} else {
      px.cleanInit()
  	}
  } else {
    if px.Persist {
    // Restore persistent state if it exists.
      px.restore(peers, me)
    } else {
      px.cleanInit()
      fmt.Println("clean init complete", px.me)
    }
  }

  // Continue with initialization as before.
  go px.MonitorDecisions()  // Monitor has mencius check in function

  if rpcs != nil {
    // caller will create socket &c
    rpcs.Register(px)
  } else {
    rpcs = rpc.NewServer()
    rpcs.Register(px)

    // prepare to receive connections from clients.
    // change "unix" to "tcp" to use over a network.
    os.Remove(peers[me]) // only needed for "unix"
    l, e := net.Listen("unix", peers[me]);
    if e != nil {
      log.Fatal("listen error: ", e);
    }
    px.l = l

    // please do not change any of the following code,
    // or do anything to subvert it.

    // create a thread to accept RPC connections
    go func() {
      for px.dead == false {
        conn, err := px.l.Accept()
        if err == nil && px.dead == false {
          if px.unreliable && (rand.Int63() % 1000) < 100 {
            // discard the request.
            conn.Close()
          } else if px.unreliable && (rand.Int63() % 1000) < 200 {
            // process the request but force discard of reply.
            c1 := conn.(*net.UnixConn)
            f, _ := c1.File()
            err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
            if err != nil {
              fmt.Printf("shutdown: %v\n", err)
            }
            px.rpcCount++
            go rpcs.ServeConn(conn)
          } else {
            px.rpcCount++
            go rpcs.ServeConn(conn)
          }
        } else if err == nil {
          conn.Close()
        }
        if err != nil && px.dead == false {
          fmt.Printf("Paxos(%v) accept: %v\n", me, err.Error())
        }
      }
    }()
  }
  if px.Mencius {
    go func() {
      for px.dead == false {
        px.tickFail()
        time.Sleep(100 * time.Millisecond)
      }
    }()
    go func() {
      for px.dead == false {
        px.tickSkip()
        time.Sleep(1000 * time.Millisecond)
      }
    }()
  }


  return px
}

// called from make after restore. holds px.mu
func (px *Paxos) resetMenciusPtr() {
  i := px.MinHelper()
  for {
    if i % len(px.peers) == px.me {
      inst := px.GetInstance(i)
      if /*inst != nil && */inst.Decided == false && inst.Proposing == false && inst.PropMax == SUGGEST_SEQ {
        px.menciusPtr = i
        break
      }
    }
    i++
    //fmt.Println(px.me, " loop in reset mencius ptr")
  }
}

func (px *Paxos) restore(peers []string, me int) {
  px.dirpath = "/tmp/6.824-ksw/paxos/" + peers[me]
  os.MkdirAll(px.dirpath, 0777)
  dir, err := os.Open(px.dirpath)
  defer dir.Close()
  if err != nil {
    log.Fatalf("Paxos.restore(): couldn't open directory\n  %v\n", err)
  }
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("Paxos.restore(): couldn't read directory\n  %v\n", err)
  }
  if px.Mencius {
    if len(filenames) == 0 {
  		if debug || px.Debug {
  			fmt.Printf("no disk found for peer %v\n", peers[me])
  		}
      px.muHD.Lock()
      px.hasDisk = false
      px.hasDiskSet = true
      px.muHD.Unlock()
      if px.tryToFetchDisk() {
        px.loadDisk()
      }
  		if debug || px.Debug {
  			fmt.Printf("clean init done\n")
  		}
    } else {
  		if debug || px.Debug {
  			fmt.Printf("loading disk for peer %v\n", peers[me])
  		}
      px.muHD.Lock()
      px.hasDisk = true
      px.hasDiskSet = true
      px.muHD.Unlock()
      px.loadDisk()
  		if debug || px.Debug {
  			fmt.Printf("loaded disk\n")
  		}
    }
  } else {
    if len(filenames) == 0 {
      if debug || px.Debug {
        fmt.Printf("clean init for peer %v\n", peers[me])
      }
      px.cleanInit()
      if debug || px.Debug {
        fmt.Printf("clean init done\n")
      }
    } else {

      if debug || px.Debug {
        fmt.Printf("loading disk for peer %v\n", peers[me])
      }
      px.loadDisk()
      if debug || px.Debug {
        fmt.Printf("loaded disk\n")
      }
    }
  }
}

func (px *Paxos) tryToFetchDisk() bool {
  var availablePeers []string
  //copy(availablePeers, px.peers[:px.me])
  //availablePeers = append(availablePeers, px.peers[(px.me + 1):]...)
  for i, v := range px.peers {
    if i != px.me {
      availablePeers = append(availablePeers, v)
    }
  }
  //fmt.Println("available peers at start of try fetch ", availablePeers, px.peers[:px.me],px.peers[(px.me + 1):])
  var maybeDiskPeers = make([]string, 0)
  for !px.dead && len(availablePeers) > 0 {
    for i := 0; i < len(availablePeers); i++ {
      peer := availablePeers[i]
      gotdisk, nodisk := px.fetchRemoteDisk(peer)
      if gotdisk {
        px.muHD.Lock()
        px.hasDisk = true
        px.hasDiskSet = true
        px.muHD.Unlock()
        return true
      }
      if !nodisk {
        maybeDiskPeers = append(maybeDiskPeers, peer)
      }
    }
    availablePeers = maybeDiskPeers
    maybeDiskPeers = make([]string, 0)
  }

  px.cleanInit()
  return false
}

func (px *Paxos) fetchRemoteDisk(server string) (bool, bool) {
  //px.muDisk.Lock()
  //defer px.muDisk.Unlock()
  px.muDisk.Lock()
  if debugDisk {
    fmt.Printf("paxos %v requesting disk from %v\n",
      px.peers[px.me], server)
  }

  args := &RequestDiskArgs{}
  args.Requestor = px.peers[px.me]
  args.DiskPath = px.dirpath
  reply := &RequestDiskReply{}
  px.muDisk.Unlock()
  ok := call(server, "Paxos.RequestDisk", args, reply)
  px.muDisk.Lock()
  defer px.muDisk.Unlock()
  if ok {
    if reply.Err == OK {

      if debugDisk {
        fmt.Printf("paxos %v got disk from server %v\n",
          px.me, server)
      }

      return true, false
    } else if reply.Err == ErrNoDisk {
      return false, true
    } else if reply.Err == ErrDead {
      return false, false
    }
  } else {

    if debugDisk {
      fmt.Printf("paxos %v couldn't get disk from server %v\n",
        px.me, server)
    }

    px.loseDisk()
    return false, false
  }
  return false, false
}

func (px *Paxos) RequestDisk(args *RequestDiskArgs, reply *RequestDiskReply) error {
  px.muDisk.Lock()
  defer px.muDisk.Unlock()
  if debugDisk {
    fmt.Printf("Request Disk handler : server %v called by %v,   hasDisk = %v   hasDiskSet = %v\n",
      px.me, args.Requestor, px.hasDisk, px.hasDiskSet)
  }

  if px.dead {
    reply.Err = ErrDead
    return nil
  }

  px.muHD.Lock()
  defer px.muHD.Unlock()

  if !px.hasDiskSet {
    reply.Err = ErrDead
    return nil
  }

  if px.hasDisk {

    if debugDisk {
      fmt.Printf("  paxos %v copying disk to server %v\n",
        px.me, args.Requestor)
    }

		//restoring := px.restoring
		//if restoring {
		//	px.muRestore.Lock()
		//}
    simulateDiskTransfer(px.dirpath, args.DiskPath)
		//if restoring {
		//	px.muRestore.Unlock()
		//}
    reply.Err = OK
  } else {

    if debugDisk {
      fmt.Printf("  shardkv %v does not have disk to offer server %v\n",
        px.me, args.Requestor)
    }

    reply.Err = ErrNoDisk
  }
  return nil
}

func simulateDiskTransfer(source string, dest string) {
  os.RemoveAll(dest)
  cmd := exec.Command("cp", "-r", source, dest)
  err := cmd.Run()
  if err != nil {
    log.Fatalf("disk transfer failed from %v to %v\n  %v\n",
      source, dest, err)
  }
}

func (px *Paxos) cleanInit() {
  px.instances = make(map[int]*Instance)
  px.seqMax = -1
  px.cleanInitDoneState()
  if px.Mencius {
    px.cleanInitMencius()
  }
}

func (px *Paxos) cleanInitDoneState() {
  px.seqMins = make([]int, len(px.peers))
  for i := range px.seqMins {
    px.seqMins[i] = -1
  }
  px.cleaned = -1
}

func (px *Paxos) cleanInitMencius() {
  px.menciusPtr = px.me
  //px.skipList = []int{}
  px.skipList = make(map[int]bool)
  px.failDetection = make(map[int]int)
}

func (px *Paxos) loadDisk() {
  px.loadDoneVector()
  px.loadSeqs()
  if px.Mencius {
    px.loadMenciusState()
  }
}

func (px *Paxos) loadDoneVector() {
//  if useNew {
	px.loadDoneVector_NEW()
//  } else {
//    px.loadDoneVector_OLD()
//  }
}


func (px *Paxos) loadDoneVector_NEW() {
  f, err := os.Open(px.dirpath + "/" + "done-vector")
  defer f.Close()
  if err != nil {
    if os.IsNotExist(err) {
      px.cleanInitDoneState()
      return
    } else {
      log.Fatalf("Paxos.loadDoneVector(): couldn't open file\n  %v\n", err)
    }
  }

  px.seqMins = make([]int, len(px.peers))
  dec := gob.NewDecoder(f)
  err = dec.Decode(&px.seqMins)
  if err != nil {
    log.Fatalf("Paxos.loadDoneVector(): error reading inst: %v\n", err)
  }
  px.cleaned = px.MinHelper() - 1
  //fmt.Printf("peer[%v] read back done vector as %v\n", px.me, px.seqMins)
}

func (px *Paxos) loadDoneVector_OLD() {
  var exist bool
  os.MkdirAll(px.dirpath, 0777)
  f, err := os.Open(px.dirpath + "/" + "done-vector")
  defer f.Close()
  if err != nil {
    if os.IsNotExist(err) {
      exist = false
    } else {
      log.Fatalf("Paxos.loadDoneVector(): couldn't open file\n  %v\n", err)
    }
  } else {
    exist = true
  }

  px.seqMins = make([]int, len(px.peers))
  if !exist {
    for i := range px.seqMins {
      px.seqMins[i] = -1
    }
  } else {
    stats, err := f.Stat()
    if err != nil {
      log.Fatalf("Paxos.loadDoneVector(): couldn't read stats\n  %v\n", err)
    }
    bytes := make([]byte, stats.Size())
    _, err = f.Read(bytes)
    if err != nil {
      log.Fatalf("Paxos.loadDoneVector(): couldn't read file\n  %v\n", err)
    }
    numbers := strings.Split(strings.TrimSpace(string(bytes)), " ")
    if len(numbers) != len(px.peers) {
				fmt.Printf("Paxos.loadDoneVector(): read file contents %v\n",
        string(bytes))
      log.Fatalf("Paxos.loadDoneVector(): have %v peers but read %v numbers\n",
        len(px.peers), len(numbers))
    }

    for i := range px.seqMins {
      seq, _ := strconv.Atoi(numbers[i])
      px.seqMins[i] = seq
    }
  }
  px.cleaned = px.MinHelper() - 1
}

func (px *Paxos) loadSeqs() {
  dir, err := os.Open(px.dirpath)
  defer dir.Close()
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("Paxos.loadSeqs(): couldn't read directory\n  %v\n", err)
  }

  px.instances = make(map[int]*Instance)
  px.seqMax = -1
  for _, fname := range filenames {
    if len(fname) > 3 && fname[0:3] == "seq" {
      seq, _ := strconv.Atoi(fname[3:])
      if seq > px.seqMax {
        px.seqMax = seq
      }
      if seq > px.cleaned {
        px.loadInstance(seq)
      } else {
        os.Remove(px.dirpath + "/" + fname)

        if debugDisk {
          fmt.Printf("removed file %v/%v\n", px.dirpath, fname)
        }

      }
    }
  }
}

func (px *Paxos) loadInstance(seq int) {
//  if useNew {
	px.loadInstance_NEW(seq)
//  } else {
  //  px.loadInstance_OLD(seq)
  //}
}

func (px *Paxos) loadInstance_NEW(seq int) {
  f, err := os.Open(px.dirpath + "/" + "seq" + strconv.Itoa(seq))
  defer f.Close()
  if err != nil {
    log.Fatalf("Paxos.loadInstance(): couldn't open file\n  %v\n", err)
  }

  inst := &Instance{}
  dec := gob.NewDecoder(f)
  err = dec.Decode(inst)
  if err != nil {
    log.Fatalf("Paxos.loadInstance(): error reading inst: %v\n", err)
  }
  px.instances[seq] = inst
  //fmt.Printf("peer[%v] read back instance[%v] as %v\n", px.me, seq, inst)
}

func (px *Paxos) loadInstance_OLD(seq int) {
  inst := px.GetInstance(seq) // should create a new *Instance
  if inst == nil {
    log.Fatalf("Paxos.loadInstance(): asking for non-existent seq%v\n", seq)
  }

  f, err := os.Open(px.dirpath + "/" + "seq" + strconv.Itoa(seq))
  defer f.Close()
  if err != nil {
    log.Fatalf("Paxos.loadInstance(): couldn't open file\n  %v\n", err)
  }
  stats, err := f.Stat()
  if err != nil {
    log.Fatalf("Paxos.loadInstance(): couldn't read file stats\n  %v\n", err)
  }
  bytes := make([]byte, stats.Size())
  _, err = f.Read(bytes)
  if err != nil {
    fmt.Printf("Paxos.loadInstance(): read file contents\n  %v\n",
      string(bytes))
    log.Fatalf("Paxos.loadInstance(): couldn't read file\n  %v\n", err)
  }
  lines := strings.Split(string(bytes), "\n")

  inst.PropMax, _ = strconv.Atoi(lines[0])
  inst.AcceptMax, _ = strconv.Atoi(lines[1])
  inst.AcceptValue = string2value(lines[2:])
}

func (px *Paxos) loadMenciusState() {
  f, err := os.Open(px.dirpath + "/" + "mencius")
  defer f.Close()
  if err != nil {
    if os.IsNotExist(err) {
      px.cleanInitMencius()
      return
    } else {
      log.Fatalf("Paxos.loadMenciusState(): couldn't open file\n  %v\n", err)
    }
  }

  var tempMenciusPtr int
  //px.skipList = []int{}
  px.skipList = make(map[int]bool)
  px.failDetection = make(map[int]int)

  dec := gob.NewDecoder(f)
  err = dec.Decode(&tempMenciusPtr)
  if err != nil {
    log.Fatalf("Paxos.loadMenciusState(): error reading mencius ptr: %v\n", err)
  }
  px.menciusPtr = tempMenciusPtr
  err = dec.Decode(&px.skipList)
  if err != nil {
    log.Fatalf("Paxos.loadMenciusState(): error reading skip list: %v\n", err)
  }
  err = dec.Decode(&px.failDetection)
  if err != nil {
    log.Fatalf("Paxos.loadMenciusState(): error reading failure map: %v\n", err)
  }
}

func (px *Paxos) loseDisk() {

  if debugDisk {
    defer fmt.Printf("cleared directory %v\n", px.dirpath)
  }

  err := os.RemoveAll(px.dirpath)
  if err != nil {
    log.Fatalf("Paxos.loseDisk(): couldn't delete directory\n  %v\n", err)
  }
  os.MkdirAll(px.dirpath, 0777)
}

func (px *Paxos) deleteOldDisk() {
  os.MkdirAll(px.dirpath, 0777)
  dir, err := os.Open(px.dirpath)
  defer dir.Close()
  if err != nil {
    log.Fatalf("Paxos.deleteOldDisk(): couldn't open directory\n  %v\n", err)
  }
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("Paxos.deleteOldDisk(): couldn't read directory\n  %v\n", err)
  }

  for _, fname := range filenames {
    if len(fname) > 3 && fname[0:3] == "seq" {
      seq, _ := strconv.Atoi(fname[3:])
      if seq <= px.cleaned {
        os.Remove(px.dirpath + "/" + fname)

        if debugDisk {
          fmt.Printf("removed file %v/%v\n", px.dirpath, fname)
        }

      }
    }
  }
}

func (px *Paxos) saveDoneState() {
//  if useNew {
	px.saveDoneState_NEW()
//  } else {
//    px.saveDoneState_OLD()
//  }
}

func (px *Paxos) saveDoneState_NEW() {

  if debugDisk {
    defer fmt.Printf("wrote %v\n", px.dirpath + "/" + "done-vector")
  }

  f, err := os.Create(px.dirpath + "/" + "done-vector")
  defer f.Close()
  if err != nil {
    log.Fatalf("Paxos.saveDoneState(): error creating file\n  %v\n", err)
  }

  enc := gob.NewEncoder(f)
  err = enc.Encode(&px.seqMins)
  if err != nil {
    log.Fatalf("Paxos.saveDoneState(): error writing done vector\n  %v\n", err)
  }
}

func (px *Paxos) saveAcceptorState(seq int) {

  if px.Persist {
  if useNew {
    px.saveAcceptorState_NEW(seq)
  } else {
    px.saveAcceptorState_OLD(seq)
  }
  }
}

func (px *Paxos) saveAcceptorState_NEW(seq int) {
  inst := px.GetInstance(seq)

  if debugDisk {
    defer fmt.Printf("wrote %v\n", px.dirpath + "/" + "seq" + strconv.Itoa(seq))
  }

  f, err := os.Create(px.dirpath + "/" + "seq" + strconv.Itoa(seq))
  defer f.Close()
  if err != nil {
    log.Fatalf("Paxos.saveAcceptorState(): error creating file\n  %v\n", err)
  }

  enc := gob.NewEncoder(f)
  err = enc.Encode(inst)
  if err != nil {
    log.Fatalf("Paxos.saveAcceptorState(): error writing inst\n  %v\n", err)
  }
}


func (px *Paxos) saveMenciusState() {
  if px.Persist {

    if debugDisk {
      defer fmt.Printf("wrote %v\n", px.dirpath + "/" + "mencius")
    }

    f, err := os.Create(px.dirpath + "/" + "mencius")
    defer f.Close()
    if err != nil {
      log.Fatalf("Paxos.saveMenciusState(): error creating file\n  %v\n", err)
    }

    enc := gob.NewEncoder(f)
    err = enc.Encode(&px.menciusPtr)
    if err != nil {
      log.Fatalf("Paxos.saveMenciusState(): error writing mencius pointer\n  %v\n", err)
    }
    err = enc.Encode(&px.skipList)
    if err != nil {
      log.Fatalf("Paxos.saveMenciusState(): error writing skip list\n  %v\n", err)
    }
    err = enc.Encode(&px.failDetection)
    if err != nil {
      log.Fatalf("Paxos.saveMenciusState(): error writing failure map\n  %v\n", err)
    }
  }
}

func (px *Paxos) GetInstances() map[int]*Instance {
	return px.instances
}

// OLD CODE

func (px *Paxos) saveDoneState_OLD() {
  var dirpath, filename string
  dirpath = px.dirpath
  filename = "done-vector"

  var data string
  for _, seq := range px.seqMins {
    data += strconv.Itoa(seq) + " "
  }
  data += "\n"

  writeFile(dirpath, filename, data)
}

func (px *Paxos) saveAcceptorState_OLD(seq int) {
  var dirpath, filename string
  dirpath = px.dirpath
  filename = "seq" + strconv.Itoa(seq)

  var data string
  inst := px.GetInstance(seq)
  data += strconv.Itoa(inst.PropMax) + "\n"
  data += strconv.Itoa(inst.AcceptMax) + "\n"
  data += value2string(inst.AcceptValue) + "\n"

  writeFile(dirpath, filename, data)
}

func value2string(value interface{}) string {
  var result string
  switch v := value.(type) {
  default:
    log.Fatalf("value2string(): can't encode value\n  value: %v\n", v)
    result = ""
  case nil:
    result = "nil"
  case int:
    result = "int\n" + strconv.Itoa(v)
  case string:
    result = "string\n" + v
  case util.Op:
    // TODO
    result = "Op"
  }
  return result
}

func string2value(lines []string) interface{} {
  var value interface{}
  switch lines[0] {
  default:
    log.Fatalf("string2value(): can't parse data\n  type: %v\n  data: %v\n",
      lines[0], lines[1:])
    value = nil
  case "nil":
    value = nil
  case "int":
    value, _ = strconv.Atoi(lines[1])
  case "string":
    value = lines[1]
  case "Op":
    // TODO
    value = NoOp
  }
  return value
}

func writeFile(dirpath string, filename string, data string) {
  if debugDisk {
    defer fmt.Printf("saved file %v/%v\n", dirpath, filename)
  }

  var err error
  var f *os.File
  os.MkdirAll(dirpath, 0777)
  f, err = os.Create(dirpath + "/" + filename)
  defer f.Close()
  if err != nil {
    log.Fatalf("writeFile(): couldn't create file\n  %v\n", err)
  }
  _, err = f.WriteString(data)
  if err != nil {
    log.Fatalf("writeFile(): couldn't write file\n  %v\n", err)
  }
  f.Sync()
}

func changeDir(dirpath string) {
  err := os.Chdir(dirpath)
  if err != nil {
    if os.IsNotExist(err) {
      err := os.MkdirAll(dirpath, 0777)
      if err != nil {
        log.Fatalf("changeDir(): couldn't create directory\n  %v\n", err)
      }
      changeDir(dirpath)
    } else {
      log.Fatalf("changeDir(): couldn't change directory\n  %v\n", err)
    }
  }
}

func (px *Paxos) ownInstance(seq int) bool {
  return (px.me == seq % len(px.peers))
}

// called from accept handler, which holds px.mu and checks mencius
func (px *Paxos) insertMySkips(seq int) {

  if seq > px.menciusPtr {
    for i := px.menciusPtr; i < seq; i+= len(px.peers) {
            //px.Start(seq, NoOp)
      inst := px.GetInstance(i)
      if inst.Proposing == false { // we haven't proposed in a slot we own so we skip
        inst.Proposing = true
        inst.Decided = true
        inst.AcceptValue = NoOp
        inst.DecideValue = NoOp

        //px.skipList = append(px.skipList, i)
        px.skipList[i] = true

        //fmt.Printf("Peer %d deciding to skip %d\n", px.me, i)
      }

    }
  }

  for ; px.menciusPtr < seq; {
    px.menciusPtr += len(px.peers)
  }

  px.muDisk.Lock()
  px.saveMenciusState()
  px.muDisk.Unlock()
}

func (px *Paxos) processSkipList(sList map[int]bool){
  if px.Mencius {
    for k,_ := range sList {
        px.skipList[k] = true
        if k < px.MinHelper(){
          delete(px.skipList, k)
        }
        //px.Start(seq, NoOp)
        inst := px.GetInstance(k)
        //inst.proposing = true
        // TODO maybe add ssertion
        //if inst != nil && inst.decided != true{
        if inst != nil && inst.Decided == false && inst.Proposing == false && inst.PropMax == SUGGEST_SEQ {
          inst.Decided = true
          inst.AcceptValue = NoOp
          inst.DecideValue = NoOp
          //fmt.Printf("Peer %d received a skip for %d\n", px.me, v)
          px.muDisk.Lock()
          px.saveAcceptorState(k)
          px.muDisk.Unlock()
        }
    }
    px.muDisk.Lock()
    px.saveMenciusState()
    px.muDisk.Unlock()
  }

}

func (px *Paxos) tickSkip(){
  // push skip list to all servers to avoid idleness causing holes
  // and to be able to prune the list
  accepted := 1 // we count as one and skips are processed before entering list
  px.mu.Lock()
  args := &RecSkipArgs{px.skipList, px.failDetection, px.me}
  px.mu.Unlock()
  for i := range px.peers {
    reply := &RecSkipReply{}
    if i == px.me {
      //px.Accept(args, reply)
    } else {
      ok := call(px.peers[i], "Paxos.RecSkip", args, reply)
      if ok && reply.OK {
        accepted += 1
      }
    }
  }
}

func (px *Paxos) tickFail(){
  //fmt.Println()
  for i := range px.peers {
    px.mu.Lock()
    if px.menciusPtr - px.failDetection[i] > ALLOWED_LAG {
      px.revokeSlots(i, px.failDetection[i])
      px.failDetection[i] += REVOKE_NUM
    }
    px.mu.Unlock()
  }

}

func (px *Paxos) RecSkip(args *RecSkipArgs, reply *RecSkipReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  if px.Mencius {
    //px.failDetection[args.Me] = args.MenciusPtr
    px.processMenciusPtrs(args.MenciusPtr)
    px.processSkipList(args.SkipList)
    reply.MenciusPtr = px.failDetection //px.menciusPtr
    reply.OK = true
  }
  return nil
}

func (px *Paxos) processMenciusPtrs(ptrs map[int]int){

  for k, v := range ptrs {
    if  px.failDetection[k] < v {
      px.failDetection[k] = v
    }
  }
  
  px.muDisk.Lock()
  px.saveMenciusState()
  px.muDisk.Unlock()

}

func (px *Paxos) revokeSlots(peer int, menciusPtr int) {
  if px.Mencius {
    for i:= menciusPtr; i < menciusPtr + REVOKE_NUM; i += len(px.peers){
      px.Start(i, NoOp)
    }
  }
}


func (px *Paxos) LearnInstance(args *LearnInstanceArgs, reply *LearnInstanceReply) error {
  px.mu.Lock()
  defer px.mu.Unlock()
  reply.Val = px.GetInstance(args.SeqNum)

  return nil
}



