package shardkv

import "fmt"
import "log"
import "math"
import "math/rand"
import "os"
import "shardmaster"
import "strings"
import "time"

var benchmarking = false

var T *log.Logger
var L *log.Logger

var t0 = time.Now()

func marktime() int64 {
  return time.Since(t0).Nanoseconds()
}

func Benchmark(scenario string, unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // Initialize logging of throughput and latency.
  benchmarking = true
  debug = false
  debugDisk = false
  persist = p != 0
  mencius = m != 0
  noMemory = c != 0

  suffix := fmt.Sprintf("-p%v-m%v-c%v", p, m, c)

  ft, _ := os.Create("throughput" + suffix + ".csv")
  defer ft.Close()
  T = log.New(ft, "", 0)

  fl, _ := os.Create("latency" + suffix + ".csv")
  defer fl.Close()
  L = log.New(fl, "", 0)

  // Run the scenario.
  switch strings.ToLower(scenario) {
  case "":
    fallthrough
  case "basic":
    ScenarioBasic(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "uniform":
    ScenarioUniform(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "focused":
    ScenarioFocused(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "failsinglewithdisk":
    ScenarioFailSingleWithDisk(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "failsinglewithoutdisk":
    ScenarioFailSingleWithoutDisk(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "failallwithdisk":
    ScenarioFailAllWithDisk(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "failallsomedisk":
    ScenarioFailAllSomeDisk(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  case "failrandom":
    ScenarioFailRandom(unreliable, ng, np, ns, nc, rps, dur, p, m, c)
  }
}

func ScenarioBasic(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // single active group
  // equally random puts and gets to a single key

  smh, gids, ha, _, clean := Setup("basic", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  for i := 0; i < nc; i++ {
    go ClientBasic(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(time.Duration(dur) * time.Second)
}

func ClientBasic(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get("a") })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put("a", "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func (ck *Clerk) timeFunc(t time.Time, reqtype string, f func()) {
  start := marktime()
  defer func() {
    end := marktime()
    elapsed := end - start
    L.Printf("%v,%v,%v,%v,%v,%v\n",
      ck.uid, ck.reqNum, reqtype, start, end, elapsed)
  }()
  f()
}

func ScenarioUniform(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // all groups active
  // equally random puts and gets to uniformly random keys

  smh, gids, ha, _, clean := Setup("uniform", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

  for i := 0; i < nc; i++ {
    go ClientUniform(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(time.Duration(dur) * time.Second)
}

func ClientUniform(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFocused(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // all groups active
  // puts distributed across the key space
  // gets on a focused subset

  smh, gids, ha, _, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

  for i := 0; i < nc; i++ {
    go ClientFocused(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(time.Duration(dur) * time.Second)
}

func ClientFocused(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      if rand.Intn(10) < 5 {
        num := int(math.Exp2(float64(rand.Intn(7))))
        key := string(byte(num))
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        key := string(byte(rand.Intn(120)))
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFailSingleWithDisk(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // single group
  // 20-second test
  // single server dies at t=5, then recovers with disk at t=10
  // puts and gets distributed uniformly

  smh, gids, ha, sa, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  dur = 20
  for i := 0; i < nc; i++ {
    go ClientFailSingleWithDisk(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(5 * time.Second)
  sa[0][0].Kill(true)
  time.Sleep(5 * time.Second)
  //sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
  sa[0][0] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 0)
  sa[0][0].unreliable = unreliable
  time.Sleep(10 * time.Second)
}

func ClientFailSingleWithDisk(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFailSingleWithoutDisk(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // single group
  // 20-second test
  // single server dies at t=5, then recovers without disk at t=10
  // puts and gets distributed uniformly

  smh, gids, ha, sa, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  dur = 20
  for i := 0; i < nc; i++ {
    go ClientFailSingleWithoutDisk(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(5 * time.Second)
  sa[0][0].Kill(false)
  time.Sleep(5 * time.Second)
  //sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
  sa[0][0] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 0)
  sa[0][0].unreliable = unreliable
  time.Sleep(10 * time.Second)
}

func ClientFailSingleWithoutDisk(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFailAllWithDisk(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // all groups
  // 20-second test
  // all servers die at t=5
  // group 0 recovers with disk at t=10
  // groups 1 and 2 recover similarly at t=11,12
  // puts and gets distributed uniformly

  smh, gids, ha, sa, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

  dur = 20
  for i := 0; i < nc; i++ {
    go ClientFailAllWithDisk(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(5 * time.Second)
  fmt.Println(sa)
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      sa[i][j].Kill(true)
    }
  }
  time.Sleep(5 * time.Second)
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
      sa[i][j].unreliable = unreliable
    }
    time.Sleep(1 * time.Second)
  }
  time.Sleep(7 * time.Second)
}

func ClientFailAllWithDisk(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFailAllSomeDisk(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // all groups
  // 20-second test
  // all servers die at t=5
  // group 0 recovers t=10, but only one peer has disk
  // groups 1 and 2 recover similarly at t=11,12
  // puts and gets distributed uniformly

  smh, gids, ha, sa, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

  dur = 20
  for i := 0; i < nc; i++ {
    go ClientFailAllWithDisk(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  time.Sleep(5 * time.Second)
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      if j == 0 {
        sa[i][j].Kill(true)
      } else {
        sa[i][j].Kill(false)
      }
    }
  }
  time.Sleep(5 * time.Second)
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
      sa[i][j].unreliable = unreliable
    }
    time.Sleep(1 * time.Second)
  }
  time.Sleep(7 * time.Second)
}

func ClientFailAllSomeDisk(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}

func ScenarioFailRandom(unreliable bool, ng, np, ns, nc, rps, dur int, p, m, c int) {
  // all groups
  // servers die randomly
  // puts and gets distributed uniformly

  smh, gids, ha, sa, clean := Setup("focused", unreliable)
  defer clean()

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

  for i := 0; i < nc; i++ {
    go ClientFailRandom(smh, rps, dur)
    time.Sleep(time.Second / time.Duration(rps) / time.Duration(nc))
  }
  
  doneChannel := make(chan bool)
  go crashRandomly(smh, gids, ha, sa, unreliable, doneChannel)
  defer func() {
    doneChannel <- true
    <-doneChannel
  }()

  time.Sleep(time.Duration(dur) * time.Second)
}

func ClientFailRandom(smh []string, rps int, dur int) {
  ck := MakeClerk(NShards, smh)
  ticker := time.NewTicker(time.Second / time.Duration(rps))
  go func () {
    for t := range ticker.C {
      key := string(byte(rand.Intn(120)))
      if rand.Intn(10) < 5 {
        ck.timeFunc(t, "Get", func() { ck.Get(key) })
      } else {
        ck.timeFunc(t, "Put", func() { ck.Put(key, "x") })
      }
    }
  }()
  time.Sleep(time.Duration(dur) * time.Second)
  ticker.Stop()
}
