package shardkv

import "runtime"
import "strconv"
import "os"
import "hash/fnv"
import "shardmaster"
import "math/rand"
import "time"

//
// Sharded key/value server.
// Lots of replica groups, each running op-at-a-time paxos.
// Shardmaster decides which group serves each shard.
// Shardmaster may change shard assignment from time to time.

var persist bool = true
var noMemory bool = false // if noMemory, then persist
var NShards = 20
var mencius = false

func hash(s string) uint32 {
  h := fnv.New32a()
  h.Write([]byte(s))
  return h.Sum32()
}

const (
  OK = "OK"
  ErrDead = "ErrDead"
  ErrNoKey = "ErrNoKey"
  ErrWrongGroup = "ErrWrongGroup"
	ErrOldRequest = "ErrOldRequest"
	ErrNotReadyYet = "ErrNotReadyYet"
  ErrNoDisk = "ErrNoDisk"
)
type Err string

type GetArgs struct {
  Key string
  UID string
  ReqNum int
}

type GetReply struct {
  Err Err
  Value string
}

type PutArgs struct {
  Key string
  Value string
  DoHash bool  // For PutHash
  UID string
  ReqNum int
}

type PutReply struct {
  Err Err
  PreviousValue string   // For PutHash
}

type ReconfigArgs struct {
	Config shardmaster.Config
}

type RequestShardArgs struct {
	ConfigNum int
	Shard int
	Loc string
}

type RequestShardReply struct {
	Table map[string]string
  Requests map[string]Req
	Err Err
}

type RequestDiskArgs struct {
  Requestor string
  DiskPath string
}

type RequestDiskReply struct {
  Err Err
}

func port(tag string, host int) string {
  s := "/var/tmp/824-"
  s += strconv.Itoa(os.Getuid()) + "/"
  os.Mkdir(s, 0777)
  s += "skv-"
  s += strconv.Itoa(os.Getpid()) + "-"
  s += tag + "-"
  s += strconv.Itoa(host)
  return s
}

func NextValue(hprev string, val string) string {
  h := hash(hprev + val)
  return strconv.Itoa(int(h))
}

func mcleanup(sma []*shardmaster.ShardMaster) {
  for i := 0; i < len(sma); i++ {
    if sma[i] != nil {
      sma[i].Kill()
    }
  }
}

func cleanup(sa [][]*ShardKV) {
  for i := 0; i < len(sa); i++ {
    for j := 0; j < len(sa[i]); j++ {
      sa[i][j].Kill(true)
    }
  }
}

func Setup(tag string, unreliable bool) ([]string, []int64, [][]string, [][]*ShardKV, func()) {
  runtime.GOMAXPROCS(4)

  const nmasters = 3
  var sma []*shardmaster.ShardMaster = make([]*shardmaster.ShardMaster, nmasters)
  var smh []string = make([]string, nmasters)
  // defer mcleanupTest(sma)
  for i := 0; i < nmasters; i++ {
    smh[i] = port(tag+"m", i)
  }
  for i := 0; i < nmasters; i++ {
    sma[i] = shardmaster.StartServer(NShards, smh, i)
  }

  const ngroups = 3   // replica groups
  const nreplicas = 3 // servers per group
  gids := make([]int64, ngroups)    // each group ID
  ha := make([][]string, ngroups)   // ShardKV ports, [group][replica]
  sa := make([][]*ShardKV, ngroups) // ShardKVs
  // defer cleanupTest(sa)
  for i := 0; i < ngroups; i++ {
    gids[i] = int64(i + 100)
    sa[i] = make([]*ShardKV, nreplicas)
    ha[i] = make([]string, nreplicas)
    for j := 0; j < nreplicas; j++ {
      ha[i][j] = port(tag+"s", (i*nreplicas)+j)
    }
    for j := 0; j < nreplicas; j++ {
      sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
      sa[i][j].unreliable = unreliable
    }
  }

  clean := func() { cleanup(sa) ; mcleanup(sma) }
  return smh, gids, ha, sa, clean
}

// randomly kills processes, and randomly deletes their disks.
// at most a minority of processes in a particular group will
// ever lose their disk
func crashRandomly(smh []string, gids []int64, ha [][]string, sa [][]*ShardKV, unreliable bool, doneChannel chan bool) {
	diskless := make([]int, len(sa))
	dead := make([]map[int]bool, len(sa))
	for g := 0; g < len(sa); g++ {
		dead[g] = make(map[int]bool)
	}
	for {
		select {
		case <-doneChannel:
			doneChannel <- true
			return
		default:
			for g := 0; g < len(sa); g++ {
				for i := 0; i < len(sa[g]); i++ {
					if rand.Intn(10) < 1 {
						if rand.Intn(10) < 1 && diskless[g] < len(sa[g]) - (len(sa[g])/2 + 1) {
							diskless[g] += 1
							sa[g][i].Kill(false)
							dead[g][i] = true
						} else {
							sa[g][i].Kill(true)
							dead[g][i] = true
						}
						time.Sleep(10 * time.Millisecond)
					}
					if dead[g][i] && rand.Intn(10) < 9 {
						dead[g][i] = false
						thisDebug := sa[g][i].Debug
						sa[g][i] = StartServer(mencius, NShards, noMemory, persist, gids[g], smh, ha[g], i)
						if unreliable {
							sa[g][i].unreliable = true
						}
						sa[g][i].Debug = thisDebug
						time.Sleep(10 * time.Millisecond)
					}
				}
			}
			time.Sleep(100 * time.Millisecond)
		}
	}
}
