package shardkv

import "io/ioutil"
import "encoding/gob"
import "fmt"
import "log"
import "math/rand"
import "net"
import "net/rpc"
import "paxos"
import "os"
import "os/exec"
import "path"
import "shardmaster"
import "strconv"
import "strings"
import "sync"
import "syscall"
import "time"
import "util"

var debug = false
var debugDisk = false

var NoOp = util.Op{Type: paxos.NoOpType}

type Req struct {
  Num int
  Val string
}

type Shard struct {
  Table map[string]string
  Requests map[string]Req
}

type ShardKV struct {
  dirpath string
  mu sync.Mutex
  muDisk sync.Mutex
  muRestore sync.Mutex
  l net.Listener
  servers []string
  me int
  dead bool // for testing
  unreliable bool // for testing
  sm *shardmaster.Clerk
  px *paxos.Paxos

  gid int64 // my replica group ID
  currentConfig shardmaster.Config
	configLock sync.Mutex
  myShards map[int]*Shard // holds shard data
  oldShards map[int]map[int]*Shard // previously transfered shards, (config ID, shard Id) -> shard
  pxMaxSeq int
	NShards int

	myLoc string // ###
	hasDisk bool
	restoring bool
	Persist bool
	noMemory bool
	mencius bool

	Debug bool
	doneOne bool
}

func (kv *ShardKV) Get(args *GetArgs, reply *GetReply) error {
	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "Trying to get", args.Key, "requested by", args.UID)
	}
	opID := "get" + "-" + args.UID + "-" + strconv.Itoa(args.ReqNum)
  op := util.Op{"Get", args, reply, opID}
  kv.submit(op)
  return nil
}

func (kv *ShardKV) Put(args *PutArgs, reply *PutReply) error {
	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "Trying to put", args.Key, "requested by", args.UID)
	}
	opID := "put" + "-" + args.UID + "-" + strconv.Itoa(args.ReqNum)
  op := util.Op{"Put", args, reply, opID}
  kv.submit(op)
  return nil
}

//
// Ask the shardmaster if there's a new configuration;
// if so, re-configure.
//
func (kv *ShardKV) tick() {

	kv.configLock.Lock()
  currentNum := kv.currentConfig.Num
	kv.configLock.Unlock()

	latestNum := kv.sm.Query(-1).Num

  if currentNum != latestNum {

		if debug || kv.Debug {
			fmt.Println(kv.me, kv.gid, kv.myLoc, "tick", currentNum, latestNum)
		}

    for i := currentNum + 1; i <= latestNum; i++ {
      config := kv.sm.Query(i)
			opID := "reconfig" + "-" + strconv.Itoa(config.Num)
      op := util.Op{"Reconfigure", ReconfigArgs{config}, nil, opID}
      kv.submit(op)
    }
  }
}

//
// Ask paxos if there are any new operations if so, apply them.
//
func (kv *ShardKV) tock() {

	kv.mu.Lock()
	defer kv.mu.Unlock()

	if !kv.dead {
		var pxMax int
		if kv.mencius {
			pxMax = kv.px.GetNextSlot() - 1
		} else {
			pxMax = kv.px.Max()
		}
		myMax := kv.pxMaxSeq
		
		if myMax != pxMax {

			if debug || kv.Debug {
				fmt.Println(kv.me, kv.gid, kv.myLoc, "tock", myMax, pxMax, kv.px.GetSeqMins())
			}
			
			kv.updateFromPx(pxMax)
		}
	}
}

func (kv *ShardKV) submit(op util.Op) {

  success := false
  for !success && !kv.dead {

    kv.mu.Lock()

		restoring := kv.restoring
		if restoring {
			kv.muRestore.Lock()
		}

    if kv.dead {
			if restoring {
				kv.muRestore.Unlock()
			}
      kv.mu.Unlock()
      break
    }

		var inst util.Op 
		if kv.mencius {
			seq := kv.px.GetNextSlot()
			kv.px.Start(seq, op)
			inst = kv.updateFromPx(seq)
			if !kv.dead {
				kv.px.Done(seq)
			}
		} else {
			kv.updateFromPx(kv.px.Max())
			
			seq := kv.pxMaxSeq + 1
			kv.pxMaxSeq = seq
			
			if debug || kv.Debug {
				fmt.Println(kv.me, kv.gid, kv.myLoc, "Submitting seq", seq, op.OpID)
			}
			
			kv.px.Start(seq, op)
			kv.waitForStatus(seq)
			inst = kv.performPxSeq(seq)
			if !kv.dead {
				kv.px.Done(seq)
			}
		}

		success = inst.OpID == op.OpID		

		if debug || kv.Debug {
			fmt.Println(kv.me, kv.gid, kv.myLoc, "tried", op.OpID, "and got", inst.OpID, "success", success)
		}

		if restoring {
			kv.muRestore.Unlock()
		}
    kv.mu.Unlock()

		time.Sleep(10 * time.Millisecond)

  }

  return
}

func (kv *ShardKV) updateFromPx(pxMax int) util.Op {

	var inst util.Op

  myMax := kv.pxMaxSeq

  num := pxMax - myMax

  if num > 0 && !kv.dead {

    doneChannel := make(chan bool, num)
    for seq := myMax + 1; seq <= pxMax; seq++ {
      go func(seq int) {
        decided,_ := kv.px.Status(seq)
        if !decided {
					if !mencius {
						kv.px.Start(seq, NoOp)
					}
          kv.waitForStatus(seq)
        }
        doneChannel <- true
      }(seq)
    }

    done := 0
    for <-doneChannel {
      done++
      if done == num {
        close(doneChannel)
        break
      }
    }

    for seq := myMax + 1; seq <= pxMax; seq++ {
      inst = kv.performPxSeq(seq)
    }

    kv.pxMaxSeq = pxMax
  }
	
	return inst
}

func (kv *ShardKV) waitForStatus(seq int) {
  to := 10 * time.Millisecond
  for !kv.dead {
		if debug || kv.Debug {
			fmt.Println(kv.servers[kv.me], kv.gid, "waiting for status", seq)
		}
    decided,_ := kv.px.Status(seq)
    if decided {
      return
    }
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
}

func (kv *ShardKV) performPxSeq(seq int) util.Op {

  _,val := kv.px.Status(seq)

  var inst util.Op
  if val != nil && !kv.dead {

    inst = val.(util.Op)

		if debug || kv.Debug {
			fmt.Println(kv.me, kv.gid, kv.myLoc, "performPxSeq executing", seq, inst.Type)
		}

    if benchmarking {
      start := marktime()
      defer func() {
        if inst.Type == paxos.NoOpType {
          return
        }
        end := marktime()
        elapsed := end - start
        T.Printf("%v,%v,%v,%v,%v,%v,%v\n",
          kv.gid, kv.me, seq, inst.Type, start, end, elapsed)
      }()
    }

    switch inst.Type {
		default:
			panic("no such operation")
    case paxos.NoOpType:
			if kv.Persist && !kv.dead {
				kv.muDisk.Lock()
				kv.writeSeq(seq)
				kv.muDisk.Unlock()
			}
    case "Get":
      switch inst.Args.(type) {
      default:
        panic("get")
      case *GetArgs:
        args := inst.Args.(*GetArgs)
        reply := inst.Reply.(*GetReply)
        reply.Value, reply.Err = kv.get(args.Key, args.UID, args.ReqNum, seq)
      case GetArgs:
        args := inst.Args.(GetArgs)
        reply := inst.Reply.(GetReply)
        reply.Value, reply.Err = kv.get(args.Key, args.UID, args.ReqNum, seq)
      }
    case "Put":
      switch inst.Args.(type) {
      default:
        panic("put")
      case *PutArgs:
        args := inst.Args.(*PutArgs)
        reply := inst.Reply.(*PutReply)
        reply.PreviousValue, reply.Err = kv.put(args.Key, args.Value, args.DoHash, args.UID, args.ReqNum, seq)
      case PutArgs:
        args := inst.Args.(PutArgs)
        reply := inst.Reply.(PutReply)
        reply.PreviousValue, reply.Err = kv.put(args.Key, args.Value, args.DoHash, args.UID, args.ReqNum, seq)
      }
    case "Reconfigure":
      switch inst.Args.(type) {
      default:
        panic("reconfigure")
      case *ReconfigArgs:
        args := inst.Args.(*ReconfigArgs)
        kv.reconfigure(args.Config, seq)
      case ReconfigArgs:
        args := inst.Args.(ReconfigArgs)
        kv.reconfigure(args.Config, seq)
      }
    }
  }

	if !kv.doneOne {
		kv.doneOne = true
	}

  return inst
}

func (kv *ShardKV) get(key string, uid string, reqNum int, seq int) (string, Err) {

  // Prepare to write seq and shard's client state.
	if kv.Persist {
		kv.muDisk.Lock()
		defer kv.muDisk.Unlock()
	}

	var val string
	var err Err

	if kv.dead {
		err = ErrDead
	}

	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "executing get", key)
	}

	var shard *Shard
	var isMyKey bool
  shardnum := key2shard(key, kv.NShards)
	if kv.noMemory {
		shard, isMyKey = kv.loadShard(shardnum)
		kv.myShards[shardnum] = shard
	} else {
		shard, isMyKey = kv.myShards[shardnum]
	}

	if isMyKey {

		lastReq, seenClient := shard.Requests[uid]

		if seenClient && lastReq.Num > reqNum  {
      err = ErrOldRequest
		} else if seenClient && lastReq.Num == reqNum {
			val = lastReq.Val
			err = OK
		} else {

			val = shard.Table[key]
      req := Req{reqNum, val}
			shard.Requests[uid] = req
			err = OK

		}
	} else {
		err = ErrWrongGroup
	}

	if kv.Persist && !kv.dead {
		if err == OK {
			kv.writeShard(shardnum)
		}
		kv.writeSeq(seq)
	}
	
	if kv.noMemory {
		delete(kv.myShards, shardnum)
		if len(kv.myShards) > 0 {
			panic("")
		}
	}

  return val, err
}


func (kv *ShardKV) put(key string, val string, doHash bool, uid string, reqNum int, seq int) (string, Err) {

  // Prepare to write seq, shard's kv pair, and shard's client state.
	if kv.Persist {
		kv.muDisk.Lock()
		defer kv.muDisk.Unlock()
	}
	
	var oldVal string
	var err Err

	if kv.dead {
		err = ErrDead
	}

	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "executing put", key)
	}

	var shard *Shard
	var isMyKey bool
  shardnum := key2shard(key, kv.NShards)
	if kv.noMemory {
		shard, isMyKey = kv.loadShard(shardnum)
		kv.myShards[shardnum] = shard
	} else {
		shard, isMyKey = kv.myShards[shardnum]
	}

	if isMyKey {

		lastReq, seenClient := shard.Requests[uid]

		if seenClient && lastReq.Num > reqNum  {
      err = ErrOldRequest
		} else if seenClient && lastReq.Num == reqNum {
			oldVal = lastReq.Val
			err = OK
		} else {

			oldVal = shard.Table[key]
			if doHash {
				val = oldVal + val
				val = strconv.Itoa(int(hash(val)))
			}

      shard.Table[key] = val
      req := Req{reqNum, oldVal}
      shard.Requests[uid] = req
			err = OK
		}		
	} else {
		err = ErrWrongGroup
	}

	if kv.Persist && !kv.dead {
		if err == OK {
			kv.writeShard(shardnum)
		}
		kv.writeSeq(seq)
	}

	if kv.noMemory {
		delete(kv.myShards, shardnum)
		if len(kv.myShards) > 0 {
			panic("")
		}
	}

  return oldVal, err
}

// figure out if i have to send/receive any shards
// can start serving shards immediately if on receiving end, but can't if sending
// can always serve shards that aren't moved
func (kv *ShardKV) reconfigure(newConfig shardmaster.Config, seq int) {

	if kv.currentConfig.Num >= newConfig.Num {
		if kv.Persist && !kv.dead {
			kv.muDisk.Lock()
			kv.writeSeq(seq)
			kv.muDisk.Unlock()		
		}
		return
	}

	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "executing reconfigure", kv.currentConfig.Num, kv.currentConfig.Shards, "to", newConfig.Num, newConfig.Shards)
	}

	if kv.currentConfig.Num != newConfig.Num - 1 {
		panic(fmt.Sprintf("reconfiguration from %v to %v", kv.currentConfig.Num, newConfig.Num))
	}

	movingShards := make(map[int]bool)
	shardsToGet := make(map[int]int64)

	for i, newGroup := range newConfig.Shards {

		oldGroup := kv.currentConfig.Shards[i]

		if oldGroup == 0 {
			if newGroup == kv.gid {
				shard := &Shard{}
				kv.myShards[i] = shard
				shard.Table = make(map[string]string)
				shard.Requests = make(map[string]Req)
        shardsToGet[i] = 0
			}
		} else {
			if oldGroup == kv.gid {
				if newConfig.Shards[i] != kv.gid {
					movingShards[i] = true
				}
			} else {
				if newConfig.Shards[i] == kv.gid {
					shardsToGet[i] = oldGroup
				}
			}
		}
	}

	if !kv.noMemory {

		kv.configLock.Lock()
		
		if len(movingShards) > 0 {
			kv.oldShards[kv.currentConfig.Num] = make(map[int]*Shard)
		}
			
		for s := range movingShards {
			shard := copyShard(kv.myShards[s])
			kv.oldShards[kv.currentConfig.Num][s] = &shard
			delete(kv.myShards, s)
		}

		kv.configLock.Unlock()
	}

  for s, g := range shardsToGet {
    if g == 0 {
      continue
    }
		success := false
		for !success && !kv.dead {
			for _, server := range kv.currentConfig.Groups[g] {
				if success {
					break
				}

				if debug || kv.Debug {
					fmt.Println(kv.me, kv.gid, kv.myLoc, "#### requesting", kv.currentConfig.Num, s, "from", server)
				}

				// the rpc sometimes doesn't time out. we could call all servers asychronously
				args := &RequestShardArgs{kv.currentConfig.Num, s, kv.myLoc}
				reply := &RequestShardReply{}
				doneChannel := make(chan bool)
				var ok bool
				var done bool
				go func() {
					ok = call(server, "ShardKV.RequestShard", args, reply)
					doneChannel <- true
				}()
				for !done {
					select{
					case <-doneChannel:
						done = true
					default:
						time.Sleep(1 * time.Millisecond)
					}
				}
				shard := Shard{reply.Table, reply.Requests}
				kv.myShards[s] = &shard
				success = ok && (reply.Err == OK)
			}
			time.Sleep(10 * time.Millisecond)
		}
  }

	if !kv.dead {
		kv.configLock.Lock()
		kv.currentConfig = newConfig
		kv.configLock.Unlock()
	}

	if kv.Persist && !kv.dead {
		// Write seq, new config, and new shards
		// Move deleted shards to old shards directory.

		kv.muDisk.Lock()
		
		if !kv.dead {
			kv.configLock.Lock()
			
			kv.writeConfig()
			for s, _ := range shardsToGet {
				kv.writeShard(s)
			}
			for s, _ := range movingShards {
				kv.moveDeletedShard(s, kv.currentConfig.Num - 1)
			}				
			kv.writeSeq(seq)

			kv.configLock.Unlock()
		}
		
		kv.muDisk.Unlock()
	}
	
	if kv.noMemory {
		for s,_ := range kv.myShards {
			delete(kv.myShards, s)
		}
		if len(kv.myShards) > 0 {
			panic("")
		}
	}

}

func (kv *ShardKV) RequestShard(args *RequestShardArgs, reply *RequestShardReply) error {

	if kv.restoring {
		kv.muRestore.Lock()
		defer kv.muRestore.Unlock()
	}

  if kv.dead {
    reply.Err = ErrDead
    return nil
  }
	
 	kv.configLock.Lock()
	defer kv.configLock.Unlock()
	
	if debug || kv.Debug {
		fmt.Println(kv.me, kv.gid, kv.myLoc, "executing RequestShard config", args.ConfigNum, "shard", args.Shard, kv.currentConfig.Num)
	}

	if args.ConfigNum < kv.currentConfig.Num {

		var shard *Shard
		var ok bool
		shardNum := args.Shard
		configNum := args.ConfigNum

		if kv.noMemory {
			shard,ok = kv.loadOldShard(configNum, shardNum)
		} else {
			shard,ok = kv.oldShards[configNum][shardNum]
		}

		if !ok {
			panic("request shard")
		}
		reply.Table = shard.Table
		reply.Requests = shard.Requests
		reply.Err = OK
	} else {
		reply.Err = ErrNotReadyYet
	}

	return nil
}

func copyShard(shard *Shard) Shard {

  table := make(map[string]string)
  requests := make(map[string]Req)

	for k,v := range shard.Table {
		table[k] = v
	}
	for k,v := range shard.Requests {
		requests[k] = v
	}

	shardCopy := Shard{table, requests}

	return shardCopy
}

// tell the server to shut itself down.
func (kv *ShardKV) Kill(keepdisk ...bool) {

  kv.dead = true

  kv.mu.Lock()
  defer kv.mu.Unlock()
	
	if kv.restoring {
		kv.muRestore.Lock()
		kv.muRestore.Unlock()
	}
	
	if kv.Persist {
		kv.muDisk.Lock()
		defer kv.muDisk.Unlock()
	}

	if debug || kv.Debug {
		if len(keepdisk) > 0 && keepdisk[0] == true {
			fmt.Printf("killed shardkv, kept disk %v\n", kv.servers[kv.me])
			fmt.Println(kv)
		} else {
			fmt.Printf("killed shardkv, lost disk %v\n", kv.servers[kv.me])
		}
	}


  kv.l.Close()
  kv.px.Kill(keepdisk...)

	if kv.Persist {
		if len(keepdisk) > 0 && keepdisk[0] == true {			
//			kv.writeSeq(kv.pxMaxSeq)
//			kv.writeConfig()
//			if !kv.noMemory {
//				for s, _ := range kv.myShards {
//					kv.writeShard(s)
//				}
//			}
		} else {
			kv.loseDisk()
		}
	}
}

//
// Start a shardkv server.
// gid is the ID of the server's replica group.
// shardmasters[] contains the ports of the
//   servers that implement the shardmaster.
// servers[] contains the ports of the servers
//   in this replica group.
// Me is the index of this server in servers[].
//
func StartServer(mencius bool, NShards int, noMemory bool, persist bool, gid int64, shardmasters []string,
                 servers []string, me int) *ShardKV {
  gob.Register(Req{})
  gob.Register(Shard{})
  gob.Register(GetArgs{})
  gob.Register(GetReply{})
  gob.Register(PutArgs{})
  gob.Register(PutReply{})
	gob.Register(ReconfigArgs{})
  gob.Register(RequestShardArgs{})
  gob.Register(RequestShardReply{})
  gob.Register(RequestDiskArgs{})
  gob.Register(RequestDiskReply{})

  kv := new(ShardKV)

  kv.dirpath = "/tmp/6.824-ksw/kvstore/" + servers[me]
  os.MkdirAll(kv.dirpath, 0777)

  kv.servers = servers
  kv.me = me
  kv.gid = gid
  kv.sm = shardmaster.MakeClerk(shardmasters)

	kv.mu = sync.Mutex{}
  kv.muDisk = sync.Mutex{}
  kv.muRestore = sync.Mutex{}
	kv.configLock = sync.Mutex{}
	kv.Persist = persist
	kv.noMemory = noMemory
	kv.NShards = NShards

	if !persist && noMemory {
		panic("start server: cannot have no memory without persisting")
	}

	if kv.Persist {
		// Restore persistent state if it exists.
		lockset := kv.reallyRestore(servers, me)
		<-lockset
	} else {
    kv.cleanInit()
	}

  // Continue with initialization as before.

  kv.myLoc = servers[me]

  rpcs := rpc.NewServer()
  rpcs.Register(kv)

	kv.mencius = mencius
  kv.px = paxos.Make(kv.mencius, kv.Persist, servers, me, rpcs)

  os.Remove(servers[me])
  l, e := net.Listen("unix", servers[me]);
  if e != nil {
    log.Fatal("listen error: ", e);
  }
  kv.l = l

  // please do not change any of the following code,
  // or do anything to subvert it.

  go func() {
    for kv.dead == false {
      conn, err := kv.l.Accept()
      if err == nil && kv.dead == false {
        if kv.unreliable && (rand.Int63() % 1000) < 100 {
          // discard the request.
          conn.Close()
        } else if kv.unreliable && (rand.Int63() % 1000) < 200 {
          // process the request but force discard of reply.
          c1 := conn.(*net.UnixConn)
          f, _ := c1.File()
          err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
          if err != nil {
            fmt.Printf("shutdown: %v\n", err)
          }
          go rpcs.ServeConn(conn)
        } else {
          go rpcs.ServeConn(conn)
        }
      } else if err == nil {
        conn.Close()
      }
      if err != nil && kv.dead == false {
        fmt.Printf("ShardKV(%v) accept: %v\n", me, err.Error())
        kv.Kill(true)
      }
    }
  }()

  go func() {
    for kv.dead == false {
			if kv.restoring {
				kv.muRestore.Lock()
				kv.muRestore.Unlock()
			}
      kv.tick()
      time.Sleep(250 * time.Millisecond)
    }
  }()

	if !kv.mencius {
		go func() {
			for kv.dead == false {
				if kv.restoring {
					kv.muRestore.Lock()
					kv.muRestore.Unlock()
				}
				kv.tock()
				time.Sleep(100 * time.Millisecond)
			}
		}()
	}

  return kv
}

func (kv *ShardKV) reallyRestore(servers []string, me int) chan bool {
  lockset := make(chan bool)
  go kv.restore(servers, me, lockset)
  return lockset
}

func (kv *ShardKV) restore(servers []string, me int, lockset chan bool) {
  kv.dirpath = "/tmp/6.824-ksw/kvstore/" + servers[me]
  os.MkdirAll(kv.dirpath, 0777)

	kv.muRestore.Lock()
	defer kv.muRestore.Unlock()
	kv.restoring = true

	kv.hasDisk = kv.findLocalDisk()

  lockset <- true

  if kv.hasDisk {
    if debugDisk || kv.Debug {
      fmt.Println("shardkv", kv.me, "found local disk")
    }
    kv.loadDisk()
    return
  }

  fmt.Println("restoring", servers[me])

  found := kv.findDisk()
  if !found {
    kv.cleanInit()
  } else {
		kv.loadDisk()
  }

	kv.restoring = false

}

func (kv *ShardKV) findDisk() bool {

  var availablePeers []string
  copy(availablePeers, kv.servers[:kv.me])
  availablePeers = append(availablePeers, kv.servers[(kv.me + 1):]...)
  var maybeDiskPeers = make([]string, 0)
  for !kv.dead && len(availablePeers) > 0 {
    for i := 0; i < len(availablePeers); i++ {
      peer := availablePeers[i]
      gotdisk, nodisk := kv.fetchRemoteDisk(peer)
      if gotdisk {
        return true
      }
      if !nodisk {
        maybeDiskPeers = append(maybeDiskPeers, peer)
      }
    }
    availablePeers = maybeDiskPeers
    maybeDiskPeers = make([]string, 0)
  }
  return false
/*
  for i, s := range kv.servers {
    if i != kv.me {

      if debugDisk || kv.Debug {
        fmt.Println("shardkv", kv.me,  "looking for remote disk", i)
      }

      if kv.fetchRemoteDisk(s) {
        return true
      }
    }
  }
  return false
*/
}

func (kv *ShardKV) findLocalDisk() bool {
  dir, err := os.Open(kv.dirpath)
  defer dir.Close()
  if err != nil {
    log.Fatalf("ShardKV.findLocalDisk(): error opening directory\n  %v\n", err)
  }
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("ShardKV.findLocalDisk(): error reading directory\n  %v\n", err)
  }
  if len(filenames) == 0 {
    return false
  } else {
    return true
  }
}
/*
func (kv *ShardKV) fetchRemoteDisk(server string) bool, bool {
  retries := 0
  for retries < 2 {
    status := kv.fetchRemoteDiskHelper(server)
    if status == true {
      return true
    }
    retries++
    time.Sleep(100 * time.Millisecond)
  }
  return false
}
*/
func (kv *ShardKV) fetchRemoteDisk(server string) (bool, bool) {

  if debugDisk || kv.Debug {
    fmt.Printf("shardkv %v requesting disk from %v\n",
      kv.servers[kv.me], server)
  }

  args := &RequestDiskArgs{}
  args.Requestor = kv.servers[kv.me]
  args.DiskPath = kv.dirpath
  reply := &RequestDiskReply{}
  ok := call(server, "ShardKV.RequestDisk", args, reply)
  if ok {
    if reply.Err == OK {

      if debugDisk || kv.Debug {
        fmt.Printf("shardkv %v got disk from server %v\n",
          kv.me, server)
      }

      return true, false
    } else if reply.Err == ErrNoDisk {
      return false, true
    } else if reply.Err == ErrDead {
      return false, false
    }
  } else {

    if debugDisk || kv.Debug {
      fmt.Printf("shardkv %v couldn't get disk from server %v\n",
        kv.me, server)
    }

    kv.loseDisk()
    return false, false
  }
  return false, false
}

func (kv *ShardKV) RequestDisk(args *RequestDiskArgs, reply *RequestDiskReply) error {
  kv.muDisk.Lock()
  defer kv.muDisk.Unlock()
  if kv.dead {
    reply.Err = ErrDead
    return nil
  }

  if kv.hasDisk {

    if debugDisk || kv.Debug {
      fmt.Printf("  shardkv %v copying disk to server %v\n",
        kv.me, args.Requestor)
    }

		restoring := kv.restoring
		if restoring {
			kv.muRestore.Lock()
		}
    simulateDiskTransfer(kv.dirpath, args.DiskPath)
		if restoring {
			kv.muRestore.Unlock()
		}
    reply.Err = OK
  } else {

    if debugDisk || kv.Debug {
      fmt.Printf("  shardkv %v does not have disk to offer server %v\n",
        kv.me, args.Requestor)
    }

    reply.Err = ErrNoDisk
  }
  return nil
}

func (kv *ShardKV) cleanInit() {

  if debugDisk || kv.Debug {
    fmt.Printf("clean init for shardkv %v\n", kv.servers[kv.me])
    defer fmt.Printf("clean initted for shardkv %v\n", kv.servers[kv.me])
  }

  kv.pxMaxSeq = -1
	kv.currentConfig = kv.sm.Query(0)
  kv.myShards = make(map[int]*Shard)
	if !kv.noMemory {
		kv.oldShards = make(map[int]map[int]*Shard)
	}

  //kv.writeConfig()
}

func (kv *ShardKV) loadDisk() {

  if debugDisk || kv.Debug {
    fmt.Printf("loading disk for shardkv %v\n", kv.servers[kv.me])
  }

  kv.loadSeq()
  kv.loadConfig()
	if kv.noMemory {
		kv.myShards = make(map[int]*Shard)
	}	else {
		kv.loadShards()
		kv.loadOldShards()
	}

	if debugDisk || kv.Debug {
		fmt.Printf("loaded disk for shardkv %v\n", kv.servers[kv.me])
		fmt.Println(kv)
	}
}

func (kv *ShardKV) loadSeq() {
  f := kv.dirpath + "/" + "seq"
	data,err := ioutil.ReadFile(f)
  if err != nil {
    log.Fatalf("ShardKV.loadSeq(): error reading file\n  %v\n", err)
  }
	kv.pxMaxSeq,err = strconv.Atoi(string(data[:]))
  if err != nil {
    log.Fatalf("ShardKV.loadSeq(): error reading seq\n  %v\n", err)
  }
}

func (kv *ShardKV) loadConfig() {
  f, err := os.Open(kv.dirpath + "/" + "config")
  defer f.Close()
  if err != nil {
    if os.IsNotExist(err) {
      kv.currentConfig = kv.sm.Query(0)
      return
    } else {
      log.Fatalf("ShardKV.loadConfig(): error opening file\n  %v\n", err)
    }
  }

  dec := gob.NewDecoder(f)
  err = dec.Decode(&kv.currentConfig)
  if err != nil {
    log.Fatalf("ShardKV.loadConfig(): error reading config: %v\n", err)
  }
}

func (kv *ShardKV) loadShards() {
  dir, err := os.Open(kv.dirpath)
  defer dir.Close()
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("ShardKV.loadShards(): error reading directory\n  %v\n", err)
  }

  kv.myShards = make(map[int]*Shard)
  for _, fname := range filenames {
    if strings.HasPrefix(fname, "shard-") {
      s, _ := strconv.Atoi(fname[6:])
      kv.myShards[s] = kv.loadShardFile(kv.dirpath, s)
    }
  }
}

func (kv *ShardKV) loadShard(s int) (*Shard, bool) {
	filename := shardPath(kv.dirpath, s)
	if _, err := os.Stat(filename); os.IsNotExist(err) {
    return nil, false
	}
	shard := kv.loadShardFile(kv.dirpath, s)
	return shard, true
}

func (kv *ShardKV) loadShardFile(dirpath string, s int) *Shard {
  f, err := os.Open(shardPath(dirpath, s))
  defer f.Close()
  if err != nil {
    log.Fatalf("ShardKV.loadShardFile(): error opening file\n  %v\n", err)
  }

  var shard Shard
  dec := gob.NewDecoder(f)
  err = dec.Decode(&shard)
  if err != nil {
    log.Fatalf("ShardKV.loadShardFile(): error reading shard: %v\n", err)
  }
  return &shard
}

func shardPath(dirpath string, s int) string {
	filename := dirpath + "/" + "shard-" + strconv.Itoa(s)	
	return filename
}

func (kv *ShardKV) loadOldShards() {
  dir, err := os.Open(kv.dirpath)
  defer dir.Close()
  dirnames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("ShardKV.loadOldShards(): error reading directory\n  %v\n", err)
  }

  kv.oldShards = make(map[int]map[int]*Shard)
  for _, dname := range dirnames {
    if strings.HasPrefix(dname, "config-") {
      c, _ := strconv.Atoi(dname[7:])
      kv.oldShards[c] = kv.loadConfigDir(c)
    }
  }
}

func (kv *ShardKV) loadOldShard(c int, s int) (*Shard, bool) {
	filename := kv.oldShardPath(c, s)
	if _, err := os.Stat(filename); os.IsNotExist(err) {
    return nil, false
	}
	shard := kv.loadShardFile(kv.oldConfigPath(c), s)
	return shard, true
}

func (kv *ShardKV) oldShardPath(c int, s int) string {
	filename := kv.oldConfigPath(c) + "/" + "shard-" + strconv.Itoa(s)	
	return filename
}

func (kv *ShardKV) oldConfigPath(c int) string {
	return kv.dirpath + "/" + "config-" + strconv.Itoa(c)
}

func (kv *ShardKV) loadConfigDir(c int) map[int]*Shard {
  configPath := kv.oldConfigPath(c)
  dir, err := os.Open(configPath)
  defer dir.Close()
  filenames, err := dir.Readdirnames(0)
  if err != nil {
    log.Fatalf("ShardKV.loadConfigDir(): error reading directory\n  %v\n", err)
  }

  var shardmap = make(map[int]*Shard)
  for _, fname := range filenames {
    if strings.HasPrefix(fname, "shard-") {
      s, _ := strconv.Atoi(fname[6:])
      shardmap[s] = kv.loadShardFile(configPath, s)
    }
  }
  return shardmap
}

func (kv *ShardKV) loseDisk() {

  if debugDisk || kv.Debug {
    fmt.Printf("losing disk for shardkv %v\n", kv.servers[kv.me])
    defer fmt.Printf("lost disk for shardkv %v\n", kv.servers[kv.me])
  }

  err := os.RemoveAll(kv.dirpath)
  if err != nil {
    log.Fatalf("error in deleting disk directory\n  %v\n", err)
  }
  os.MkdirAll(kv.dirpath, 0777)
}

func (kv *ShardKV) writeSeq(seq int) {

	kv.hasDisk = true

  if debugDisk || kv.Debug {
    defer fmt.Printf("wrote %v\n", kv.dirpath + "/" + "seq")
  }
	
  f, err := os.Create(kv.dirpath + "/" + "seq")
  defer f.Close()
	defer f.Sync()
  if err != nil {
    log.Fatalf("ShardKV.writeSeq(): error creating file\n  %v\n", err)
  }
	
  _,err = f.Write([]byte(strconv.Itoa(seq)))
  if err != nil {
    log.Fatalf("ShardKV.writeSeq(): error writing seq\n  %v\n", err)
  }
}

func (kv *ShardKV) writeConfig() {

	kv.hasDisk = true

  if debugDisk || kv.Debug {
    defer fmt.Printf("wrote %v\n", kv.dirpath + "/" + "config")
  }

  f, err := os.Create(kv.dirpath + "/" + "config")
  defer f.Close()
	defer f.Sync()
  if err != nil {
    log.Fatalf("ShardKV.writeConfig(): error creating file\n  %v\n", err)
  }

  enc := gob.NewEncoder(f)
  err = enc.Encode(&kv.currentConfig)
  if err != nil {
    log.Fatalf("ShardKV.writeConfig(): error writing config\n  %v\n", err)
  }
}

func (kv *ShardKV) writeShard(shardnum int) {

	kv.hasDisk = true

  if debugDisk || kv.Debug {
    defer fmt.Printf("wrote %v\n", kv.dirpath + "/" + "shard-" + strconv.Itoa(shardnum))
  }

  f, err := os.Create(kv.dirpath + "/" + "shard-" + strconv.Itoa(shardnum))
  defer f.Close()
	defer f.Sync()
  if err != nil {
    log.Fatalf("ShardKV.writeShard(): error creating file\n  %v\n", err)
  }

  shard, ok := kv.myShards[shardnum]
  if !ok {
    log.Fatalf("shard %v does not exist for writing\n", shardnum)
  }
  enc := gob.NewEncoder(f)
  err = enc.Encode(shard)
  if err != nil {
    log.Fatalf("ShardKV.writeShard(): error writing shard\n  %v\n", err)
  }
/*
  for k, v := range shard.Table {
    kv.writeKVEntry(shardnum, k, v)
  }
  for uid, req := range shard.Requests {
    kv.writeCkEntry(shardnum, uid, req)
  }
*/
}

func (kv *ShardKV) moveDeletedShard(shardnum int, confignum int) {
  shardFileName := "shard-" + strconv.Itoa(shardnum)
  configDirName := "config-" + strconv.Itoa(confignum)
  oldPath := kv.dirpath + "/" + shardFileName
  newPath := kv.dirpath + "/" + configDirName + "/" + shardFileName
  moveFile(oldPath, newPath)
}

func moveFile(oldpath string, newpath string) {

  if debugDisk {
    defer fmt.Printf("moved file\n  from %v\n  to %v\n", oldpath, newpath)
  }

  newdir := path.Dir(newpath)
  os.MkdirAll(newdir, 0777)

  err := os.Rename(oldpath, newpath)
  if err != nil {
    log.Fatalf("moveFile(): error moving file\n  %v\n", err)
  }
}

func simulateDiskTransfer(source string, dest string) {
  os.RemoveAll(dest)
  cmd := exec.Command("cp", "-r", source, dest)
  err := cmd.Run()
  if err != nil {
    log.Fatalf("disk transfer failed from %v to %v\n  %v\n",
      source, dest, err)
  }
}

