package shardkv

import rdebug "runtime/debug"
import "testing"
import "shardmaster"
import "runtime"
import "strconv"
import "os"
import "time"
import "fmt"
import "sync"
import "math/rand"

var includeRandomCrashTests = false
var loseDiskAtRandom = true
var includeTestBig = false

func TestBasic(t *testing.T) {
  smh, gids, ha, _, clean := Setup("basic", false)
  defer clean()

  fmt.Printf("Test: Basic Join/Leave ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(NShards, smh)

  ck.Put("a", "x")
  v := ck.PutHash("a", "b")
  if v != "x" {
    t.Fatalf("Puthash got wrong value")
  }
  ov := NextValue("x", "b")
  if ck.Get("a") != ov {
    t.Fatalf("Get got wrong value")
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  // are keys still there after joins?
  for g := 1; g < len(gids); g++ {
    mck.Join(gids[g], ha[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("joining; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  // are keys still there after leaves?
  for g := 0; g < len(gids)-1; g++ {
    mck.Leave(gids[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("leaving; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  fmt.Printf("  ... Passed\n")
}

func TestMove(t *testing.T) {
	Move(t, false)
}

func Move(t *testing.T, crashes bool) {
	tag := ""
	if crashes {
		tag = "movecrashes"
	} else {
		tag = "move"
	}
	smh, gids, ha, sa, clean := Setup(tag, false)
  defer clean()

	if crashes {
		fmt.Printf("Test: Shards really move with crashes during reconfiguration...\n")
	} else {
		fmt.Printf("Test: Shards really move ...\n")
	}

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(NShards, smh)

  // insert one key per shard
  for i := 0; i < NShards; i++ {
    ck.Put(string('0'+i), string('0'+i))
  }

  // add group 1.
  mck.Join(gids[1], ha[1])
	if crashes {
		time.Sleep(time.Duration(rand.Intn(10)) * time.Millisecond)
		for i := 0; i < len(sa); i++ {
			sa[i][0].Kill(false)
			for j := 1; j < len(sa[i]); j++ {
				sa[i][j].Kill(true)
			}
		}
		for i := 0; i < len(sa); i++ {
			for j := 0; j < len(sa[i]); j++ {
				sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
			}
		}
	}
  time.Sleep(5 * time.Second)

  // check that keys are still there.
  for i := 0; i < NShards; i++ {
    if ck.Get(string('0'+i)) != string('0'+i) {
      t.Fatalf("missing key/value")
    }
  }

  // remove sockets from group 0.
  for i := 0; i < len(ha[0]); i++ {
    os.Remove(ha[0][i])
  }

  count := 0
  var mu sync.Mutex
  for i := 0; i < NShards; i++ {
    go func(me int) {
      myck := MakeClerk(NShards, smh)
      v := myck.Get(string('0'+me))
      if v == string('0'+me) {
        mu.Lock()
        count++
        mu.Unlock()
      } else {
        t.Fatalf("Get(%v) yielded %v\n", i, v)
      }
    }(i)
  }

  time.Sleep(10 * time.Second)

  if count > NShards / 3 && count < 2*(NShards/3) {
    fmt.Printf("  ... Passed\n")
  } else {
    t.Fatalf("%v keys worked after killing 1/2 of groups; wanted %v",
      count, NShards / 2)
  }
}

func TestLimp(t *testing.T) {
  smh, gids, ha, sa, clean := Setup("limp", false)
  defer clean()

  fmt.Printf("Test: Reconfiguration with some dead replicas ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(NShards, smh)

  ck.Put("a", "b")
  if ck.Get("a") != "b" {
    t.Fatalf("got wrong value")
  }

  for g := 0; g < len(sa); g++ {
    sa[g][rand.Int() % len(sa[g])].Kill(true)
  }

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  // are keys still there after joins?
  for g := 1; g < len(gids); g++ {
    mck.Join(gids[g], ha[g])
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("joining; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  // are keys still there after leaves?
  for g := 0; g < len(gids)-1; g++ {
    mck.Leave(gids[g])
    time.Sleep(2 * time.Second)
    for i := 0; i < len(sa[g]); i++ {
      sa[g][i].Kill(true)
    }
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("leaving; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  fmt.Printf("  ... Passed\n")
}

func doConcurrent(t *testing.T, unreliable bool, crashes bool) {
	rand.Seed( time.Now().UTC().UnixNano())
  smh, gids, ha, sa, clean := Setup("conc"+strconv.FormatBool(unreliable)+strconv.FormatBool(crashes), unreliable)

	sa[2][2].Debug = true

  mck := shardmaster.MakeClerk(smh)

	if crashes {
		doneChannel := make(chan bool)
		go crashRandomly(smh, gids, ha, sa, unreliable, doneChannel)
		defer func() {
			doneChannel <- true
			<-doneChannel
//			panic("")
			clean()
		}()
	} else {
		defer clean()
	}

  for i := 0; i < len(gids); i++ {
    mck.Join(gids[i], ha[i])
  }

  const npara = 11
  var ca [npara]chan bool
  for i := 0; i < npara; i++ {
    ca[i] = make(chan bool)
    go func(me int) {
      ok := true
      defer func() { ca[me] <- ok }()
      ck := MakeClerk(NShards, smh)
      mymck := shardmaster.MakeClerk(smh)
      key := strconv.Itoa(me)
      last := ""
      for iters := 0; iters < 3; iters++ {
        nv := strconv.Itoa(rand.Int())
        v := ck.PutHash(key, nv)
        if v != last {
          ok = false
          t.Fatalf("PutHash(%v) expected %v got %v\n", key, last, v)
        }
        last = NextValue(last, nv)
        v = ck.Get(key)
        if v != last {
          ok = false
          t.Fatalf("Get(%v) expected %v got %v\n", key, last, v)
        }

        mymck.Move(rand.Int() % NShards,
          gids[rand.Int() % len(gids)])

        time.Sleep(time.Duration(rand.Int() % 30) * time.Millisecond)
      }
    }(i)
  }

  for i := 0; i < npara; i++ {
    x := <- ca[i]
    if x == false {
      t.Fatalf("something is wrong")
    }
  }
}

func TestConcurrent(t *testing.T) {
  fmt.Printf("Test: Concurrent Put/Get/Move ...\n")
  doConcurrent(t, false, false)
  fmt.Printf("  ... Passed\n")
}

func TestConcurrentUnreliable(t *testing.T) {
  fmt.Printf("Test: Concurrent Put/Get/Move (unreliable) ...\n")
  doConcurrent(t, true, false)
  fmt.Printf("  ... Passed\n")
}


func TestBasicRestarts(t *testing.T) {

	if !persist {
		return
	}

  smh, gids, ha, sa, clean := Setup("basiccrashes", false)
  defer clean()

  fmt.Printf("Test: Basic Join/Leave with crashes ...\n")
	
  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(NShards, smh)

  fmt.Println("hello 0")

	// does service keep key around after full crash?
  ck.Put("a", "x")

  fmt.Println("hello 1")

	fmt.Println(sa[0][0])
	fmt.Println(sa[0][1])
	fmt.Println(sa[0][2])

	sa[0][0].Kill(false)
	for j := 1; j < len(sa[0]); j++ {
		sa[0][j].Kill(true)
	}
	for j := 0; j < len(sa[0]); j++ {
		sa[0][j] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], j)
	}

  fmt.Println("hello 2")

	fmt.Println(sa[0][0])
	fmt.Println(sa[0][1])
	fmt.Println(sa[0][2])

  v := ck.PutHash("a", "b")
  if v != "x" {
    t.Fatalf("Puthash got wrong value")
  }
  ov := NextValue("x", "b")
  if ck.Get("a") != ov {
    t.Fatalf("Get got wrong value")
  }

  fmt.Println("hello 3")

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  fmt.Println("hello 4")

  // are keys still there if groups fail during reconfigurations from join?
  for g := 1; g < len(gids); g++ {

    fmt.Println("hello 4.0")

    mck.Join(gids[g], ha[g])
		sa[0][0].Kill(false)
		for j := 1; j < len(sa[0]); j++ {
			sa[0][j].Kill(true)
		}
		sa[g][0].Kill(false)
		for j := 1; j < len(sa[g]); j++ {
			sa[g][j].Kill(true)
		}

    fmt.Println("hello 4.1")

    time.Sleep(10 * time.Millisecond)
		for j := 0; j < len(sa[0]); j++ {
			sa[0][j] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], j)
		}
		for j := 0; j < len(sa[g]); j++ {
			sa[g][j] = StartServer(mencius, NShards, noMemory, persist, gids[g], smh, ha[g], j)
		}

    fmt.Println("hello 4.2")

    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {

      fmt.Println("hello 4.2.0")

      v := ck.Get(keys[i])

      fmt.Println("hello 4.2.1")

      if v != vals[i] {
        t.Fatalf("joining; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])

      fmt.Println("hello 4.2.2")
    }

    fmt.Println("hello 4.3")

}

  fmt.Println("hello 5")

	// are keys still there after full crash?
	for i := 0; i < len(sa); i++ {
		sa[i][0].Kill(false)
		for j := 1; j < len(sa[i]); j++ {
			sa[i][j].Kill(true)
		}
	}
	for i := 0; i < len(sa); i++ {
		for j := 0; j < len(sa[i]); j++ {
			sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
		}
	}

  fmt.Println("hello 6")

	for i := 0; i < len(keys); i++ {
		v := ck.Get(keys[i])
		if v != vals[i] {
			t.Fatalf("joining; wrong value; k=%v wanted=%v got=%v", keys[i], vals[i], v)
		}
	}

  fmt.Println("hello 7")

  // are keys still there if everyone fails after a reconfigure from a leave?
  for g := 0; g < len(gids)-1; g++ {
    mck.Leave(gids[g])
		for i := 0; i < len(sa); i++ {
			sa[i][0].Kill(false)
			for j := 1; j < len(sa[i]); j++ {
				sa[i][j].Kill(true)
			}
		}
		for i := 0; i < len(sa); i++ {
			for j := 0; j < len(sa[i]); j++ {
				sa[i][j] = StartServer(mencius, NShards, noMemory, persist, gids[i], smh, ha[i], j)
			}
		}
    time.Sleep(1 * time.Second)
    for i := 0; i < len(keys); i++ {
      v := ck.Get(keys[i])
      if v != vals[i] {
        t.Fatalf("leaving; wrong value; g=%v k=%v wanted=%v got=%v",
          g, keys[i], vals[i], v)
      }
      vals[i] = strconv.Itoa(rand.Int())
      ck.Put(keys[i], vals[i])
    }
  }

  fmt.Println("hello 8")

  fmt.Printf("  ... Passed\n")
}

func TestRecovery(t *testing.T) {

	if !persist {
		return
	}

  smh, gids, ha, sa, clean := Setup("recover", false)
  defer clean()

  fmt.Printf("Test: Failed servers get caught up ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])

  ck := MakeClerk(NShards, smh)

	// server that dies and loses memory gets caught up
	sa[0][0].Kill(true)

  ck.Put("a", "x")
	ck.PutHash("a", "b")
	time.Sleep(10 * time.Millisecond)

	sa[0][0] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 0)

	time.Sleep(1 * time.Second)

	sa[0][1].Kill(true)
	sa[0][2].Kill(true)

	sa[0][1] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 1)

	time.Sleep(1 * time.Second)

  ov := NextValue("x", "b")
  if ck.Get("a") != ov {
    t.Fatalf("Get got wrong value")
  }

	sa[0][2] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 2)

	time.Sleep(1 * time.Second)

	// server that dies and loses disk gets caught up
	sa[0][0].Kill(false)

  ck.PutHash("a", "c")
	time.Sleep(1 * time.Millisecond)

	sa[0][0] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 0)
	time.Sleep(1 * time.Second)

	sa[0][1].Kill(true)
	sa[0][2].Kill(true)

	sa[0][1] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 1)
	time.Sleep(1 * time.Second)

  ov = NextValue(ov, "c")
  if ck.Get("a") != ov {
    t.Fatalf("Get got wrong value")
  }

	sa[0][2] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 2)

	// restarted servers get caught up after joins
	sa[0][2].Kill(true)
	sa[1][2].Kill(true)
	sa[2][2].Kill(true)

  keys := make([]string, 10)
  vals := make([]string, len(keys))
  for i := 0; i < len(keys); i++ {
    keys[i] = strconv.Itoa(rand.Int())
    vals[i] = strconv.Itoa(rand.Int())
    ck.Put(keys[i], vals[i])
  }

  for g := 1; g < len(gids); g++ {
    mck.Join(gids[g], ha[g])
    time.Sleep(1 * time.Second)
  }

	// don't get keys while too many replicas are dead
	sa[0][0].Kill(true)
	sa[1][0].Kill(true)
	sa[2][0].Kill(true)

	time.Sleep(10 * time.Millisecond)

	valueChannel := make(chan string)
	for i := 0; i < len(keys); i++ {
		go func(i int) {
			valueChannel <- ck.Get(keys[i])
		}(i)
	}

	time.Sleep(1 * time.Second)

	select {
	case <-valueChannel:
		t.Fatalf("get served with too few processes alive")
	default:
		break
	}

	// get keys after replicas return
	sa[0][2] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 2)
	sa[1][2] = StartServer(mencius, NShards, noMemory, persist, gids[1], smh, ha[1], 2)
	sa[2][2] = StartServer(mencius, NShards, noMemory, persist, gids[2], smh, ha[2], 2)

	time.Sleep(1 * time.Second)

	// 2 and 1 put some new values

	sa[0][1].Kill(true)
	sa[1][1].Kill(true)
	sa[2][1].Kill(true)

	sa[0][0] = StartServer(mencius, NShards, noMemory, persist, gids[0], smh, ha[0], 0)
	sa[1][0] = StartServer(mencius, NShards, noMemory, persist, gids[1], smh, ha[1], 0)
	sa[2][0] = StartServer(mencius, NShards, noMemory, persist, gids[2], smh, ha[2], 0)

	time.Sleep(1 * time.Second)

	for i := 0; i < len(keys); i++ {
		v := ck.Get(keys[i])
		if v != vals[i] {
			t.Fatalf("joining; wrong value; k=%v wanted=%v got=%v", keys[i], vals[i], v)
		}
	}

  fmt.Printf("  ... Passed\n")
}

func TestMoveCrashes(t *testing.T) {

	if !persist {
		return
	}

	Move(t, true)
}

func TestConcurrentCrashes(t *testing.T) {

	if !persist || !includeRandomCrashTests {
		return
	}

  fmt.Printf("Test: Concurrent Put/Get/Move with crashes ...\n")
  doConcurrent(t, false, true)
  fmt.Printf("  ... Passed\n")
}

func TestConcurrentUnreliableCrashes(t *testing.T) {

	if !persist || !includeRandomCrashTests {
		return
	}

  fmt.Printf("Test: Concurrent Put/Get/Move (unreliable) with crashes ...\n")
  doConcurrent(t, true, true)
  fmt.Printf("  ... Passed\n")
}

func TestBig(t *testing.T) {

	if !includeTestBig {
		return
	}

	NShards = 100
	noMemory = true

  smh, gids, ha, sa, clean := Setup("big", false)
  defer clean()

  fmt.Printf("Test: Lots of data ...\n")

  mck := shardmaster.MakeClerk(smh)
  mck.Join(gids[0], ha[0])
  mck.Join(gids[1], ha[1])
  mck.Join(gids[2], ha[2])

	time.Sleep(1 * time.Second)

  ck := MakeClerk(NShards, smh)

	// get the current memory, putting keys after this shouldn't change
	// memory much

  runtime.GC()
  var m0 runtime.MemStats
  runtime.ReadMemStats(&m0)
	
	// put a bunch of other keys

	putStuff(ck)

	for i := 0; i < len(sa); i++ {
		for j := 0; j < len(sa[i]); j++ {
			fmt.Println(sa[i][j])
		}
	}
	
	time.Sleep(1 * time.Second)

	// check that memory didn't increase too much

  runtime.GC()
	rdebug.FreeOSMemory()
  var m1 runtime.MemStats
  runtime.ReadMemStats(&m1)

	fmt.Println(m0.Alloc, m1.Alloc)

	clean()
	
	time.Sleep(1 * time.Second)

  runtime.GC()
	rdebug.FreeOSMemory()
  runtime.ReadMemStats(&m1)

	fmt.Println(m0.Alloc, m1.Alloc)

 if m1.Alloc > m0.Alloc*10 {
 	for g := 0; g < len(sa); g++ {
 		for i := 0; i < len(sa[g]); i++ {
 			fmt.Println(sa[g][i])
 		}
 	}		
   t.Fatalf("initial puts: memory use did not shrink enough, was: %v, now: %v", m0.Alloc, m1.Alloc)
 }

	// // make sure still correct and no memory leaks after join

	// mck.Join(gids[2], ha[2])
	// time.Sleep(1 * time.Second)

	// for i := 0; i < len(keys); i++ {
	// 	v := ck.Get(keys[i])
	// 	if v != vals[i] {
	// 		t.Fatalf("after join: wrong value")
	// 	}
  //   big := make([]byte, valSize)
  //   for j := 0; j < len(big); j++ {
  //     big[j] = byte('a' + rand.Int() % 26)
  //   }
  //   ck.Put(keys[i], string(big))
	// 	vals[i] = string(big)
	// }

  // runtime.GC()
  // var m2 runtime.MemStats
  // runtime.ReadMemStats(&m2)

  // if m2.Alloc > m0.Alloc*2 {
  //   t.Fatalf("after join: memory use did not shrink enough, was: %v, now: %v", m0.Alloc, m2.Alloc)
  // }

	// // make sure still correct and no memory leaks after leave

	// mck.Leave(gids[0])
	// time.Sleep(1 * time.Second)

	// for i := 0; i < len(keys); i++ {
	// 	v := ck.Get(keys[i])
	// 	if v != vals[i] {
	// 		t.Fatalf("after leave: wrong value")
	// 	}
	// }

  // runtime.GC()
  // var m3 runtime.MemStats
  // runtime.ReadMemStats(&m3)

  // if m3.Alloc > m0.Alloc*2 {
  //   t.Fatalf("after leave: memory use did not shrink enough, was: %v, now: %v", m0.Alloc, m3.Alloc)
  // }

	// fmt.Println(m0.Alloc)
	// fmt.Println(m1.Alloc)
	// fmt.Println(m2.Alloc)
	// fmt.Println(m3.Alloc)

  fmt.Printf("  ... Passed\n")
}


func putStuff(ck *Clerk) {
	
	nOtherKeys := 5000
	valSize := 10000

	doneChannel := make(chan bool)
  for i := 0; i < nOtherKeys; i++ {
		go func() {
			key := strconv.Itoa(rand.Int())
			big := make([]byte, valSize)
			for j := 0; j < len(big); j++ {
				big[j] = byte('a' + rand.Int() % 26)
			}
			ck.Put(key, string(big))
			doneChannel<-true
		}()
  }

	done := 0
	for done < nOtherKeys {
		select {
		case <-doneChannel:
			done++
		}
		if done % 1000 == 0 {
			fmt.Println("done with putting", done)
		}
	}

	close(doneChannel)
}
