package shardmaster

import "net"
import "fmt"
import "net/rpc"
import "log"
import "paxos"
import "sync"
import "os"
import "syscall"
import "encoding/gob"
import "math/rand"
import "time"
import "util"

type ShardMaster struct {
  mu sync.Mutex
  l net.Listener
  me int
  dead bool // for testing
  unreliable bool // for testing
  px *paxos.Paxos

  configs []Config // indexed by config num

  groups []int64
  assignments map[int64]map[int]bool
  pxMaxSeq int

	NShards int
}

var NoOp = util.Op{Type: paxos.NoOpType}

const debug bool = false

func (sm *ShardMaster) Join(args *JoinArgs, reply *JoinReply) error {
  op := util.Op{"Join", args, reply, ""}
  sm.submit(op)
  return nil
}

func (sm *ShardMaster) Leave(args *LeaveArgs, reply *LeaveReply) error {
  op := util.Op{"Leave", args, reply, ""}
  sm.submit(op)
  return nil
}

func (sm *ShardMaster) Move(args *MoveArgs, reply *MoveReply) error {
  op := util.Op{"Move", args, reply, ""}
  sm.submit(op)
  return nil
}

func (sm *ShardMaster) Query(args *QueryArgs, reply *QueryReply) error {
  op := util.Op{"Query", args, reply, ""}
  sm.submit(op)
  return nil
}

func (sm *ShardMaster) submit(op util.Op) {

  success := false
  for !success && !sm.dead {

    sm.mu.Lock()

    sm.updateFromPx()

    seq := sm.pxMaxSeq + 1
    sm.pxMaxSeq = seq

    sm.px.Start(seq, op)
    sm.waitForStatus(seq)
    inst := sm.performPxSeq(seq)
    sm.px.Done(seq)

    success = inst == op

    sm.mu.Unlock()
  }

  return
}

func (sm *ShardMaster) updateFromPx() {

  pxMax := sm.px.Max()
  myMax := sm.pxMaxSeq
  num := pxMax - myMax

  if num > 0 {

    doneChannel := make(chan bool, num)
    for seq := myMax + 1; seq <= pxMax; seq++ {
      go func(seq int) {
        decided,_ := sm.px.Status(seq)
        if !decided {
          sm.px.Start(seq, NoOp)
          sm.waitForStatus(seq)
        }
        doneChannel <- true
      }(seq)
    }

    done := 0
    for <-doneChannel {
      done++
      if done == num {
        close(doneChannel)
        break
      }
    }

    for seq := myMax + 1; seq <= pxMax; seq++ {
      sm.performPxSeq(seq)
    }

    sm.pxMaxSeq = pxMax
  }
}

func (sm *ShardMaster) performPxSeq(seq int) util.Op {

  _,val := sm.px.Status(seq)

  var inst util.Op
  if val != nil {

    inst = val.(util.Op)

    switch inst.Type {
		default:
			panic("no such operation!")
    case paxos.NoOpType:
    case "Join":
      switch inst.Args.(type) {
      default:
        panic("join")
      case *JoinArgs:
        args := inst.Args.(*JoinArgs)
        if debug {
          fmt.Println(sm.me, "join", args.GID)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.join(args.GID, args.Servers)
      case JoinArgs:
        args := inst.Args.(JoinArgs)
        if debug {
          fmt.Println(sm.me, "join", args.GID)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.join(args.GID, args.Servers)
      }
    case "Leave":
      switch inst.Args.(type) {
      default:
        panic("leave")
      case *LeaveArgs:
        gid := inst.Args.(*LeaveArgs).GID
        if debug {
          fmt.Println(sm.me, "leave", gid)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.leave(gid)
      case LeaveArgs:
        gid := inst.Args.(LeaveArgs).GID
        if debug {
          fmt.Println(sm.me, "leave", gid)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.leave(gid)
      }
    case "Move":
      switch inst.Args.(type) {
      default:
        panic("move")
      case *MoveArgs:
        args := inst.Args.(*MoveArgs)
        if debug {
          fmt.Println(sm.me, "move", args.Shard, args.GID)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.move(args.Shard, args.GID)
      case MoveArgs:
        args := inst.Args.(MoveArgs)
        if debug {
          fmt.Println(sm.me, "move", args.Shard, args.GID)
          fmt.Println(sm.me, sm.query(-1), sm.assignments)
        }
        sm.move(args.Shard, args.GID)
      }
    case "Query":
      switch inst.Args.(type) {
      default:
        panic("query")
      case *QueryArgs:
        reply := inst.Reply.(*QueryReply)
        reply.Config = sm.query(inst.Args.(*QueryArgs).Num)
      case QueryArgs:
        reply := inst.Reply.(QueryReply)
        reply.Config = sm.query(inst.Args.(QueryArgs).Num)
      }
      if debug {
        fmt.Println(sm.me, "query")
      }
    }
  }

  if debug {
    fmt.Println(sm.me, sm.query(-1), sm.assignments)
  }

  return inst
}

// replica group gid is joining, give it some shards
func (sm *ShardMaster) join(gid int64, servers []string) {

  if _,exists := sm.assignments[gid]; exists {
    return
  }

  sm.groups = append(sm.groups, gid)
  sm.addNextConfig()
  config := &sm.configs[len(sm.configs) - 1]
  config.Groups[gid] = servers
  nGroups := len(config.Groups)
  minLoad := sm.NShards/nGroups

  sm.assignments[gid] = map[int]bool{}

  if nGroups == 1 {
    for i := range config.Shards {
      sm.assignments[gid][i] = true
      config.Shards[i] = gid
    }
  } else {
    overloaded := sm.whichToOverload(0)
    shards := map[int]bool{} // turn into array
    sm.balance(minLoad, overloaded, shards)
  }
}

// replica group gid is retiring, hand off all its shards
func (sm *ShardMaster) leave(gid int64) {

  if _,exists := sm.assignments[gid]; !exists {
    return
  }

  var i int
  var g int64
  for i,g = range sm.groups {
    if g == gid {
      break
    }
  }
  sm.groups = append(sm.groups[:i], sm.groups[i+1:]...)
  sm.addNextConfig()
  config := &sm.configs[len(sm.configs) - 1]
  delete(config.Groups, gid)
  shards := sm.assignments[gid] // turn into array
  nGroups := len(config.Groups)
  minLoad := sm.NShards/nGroups

  if nGroups == 0 {
    for i := range config.Shards {
      config.Shards[i] = 0
    }
  } else {
    overloaded := sm.whichToOverload(gid)
    sm.balance(minLoad, overloaded, shards)
  }
  delete(sm.assignments, gid)
}

func (sm *ShardMaster) whichToOverload(removed int64) map[int64]int {

  nGroups := len(sm.groups)
  minLoad := sm.NShards/nGroups

  nOverload := sm.NShards - minLoad*nGroups
  overloaded := make(map[int64]int, nOverload)

  // this could be faster using sorting
  n := 0
  for _,g := range sm.groups {
    if n == nOverload {
      break
    }

    if len(sm.assignments[g]) > minLoad {
      overloaded[g] = 1
      n += 1
    }
  }

  for _,g := range sm.groups {
    if n == nOverload {
      break
    }
    if overloaded[g] == 0 {
      overloaded[g] = 1
      n += 1
    }
  }

  return overloaded
}

func (sm *ShardMaster) balance(minLoad int, overloaded map[int64]int, shards map[int]bool) {

  config := &sm.configs[len(sm.configs) - 1]

  for _,g := range sm.groups {

    targetLoad := minLoad + overloaded[g]

    if len(sm.assignments[g]) > targetLoad {

      for s,_ := range config.Shards {
        if len(sm.assignments[g]) == targetLoad {
          break
        }
        if sm.assignments[g][s] {
          delete(sm.assignments[g], s)
          shards[s] = true
        }
      }
    }
  }

  for _,g := range sm.groups {

    targetLoad := minLoad + overloaded[g]

    for s,_ := range config.Shards {
      if len(sm.assignments[g]) == targetLoad {
        break
      }
      if shards[s] {
        delete(shards, s)
        sm.assignments[g][s] = true
        config.Shards[s] = g
      }
    }
  }

  if len(shards) > 0 {
    panic("lost some shards!")
  }
}

// hand off one shard from current owner to gid
func (sm *ShardMaster) move(shard int, gid int64) {

  sm.addNextConfig()
  config := &sm.configs[len(sm.configs) - 1]

  delete(sm.assignments[config.Shards[shard]], shard)

  config.Shards[shard] = gid
  sm.assignments[gid][shard] = true
}

// fetch Config # num, or latest config if num==-1
func (sm *ShardMaster) query(num int) Config {
  if num == -1 {
    return sm.configs[len(sm.configs) - 1]
  } else {
    return sm.configs[num]
  }
}


func (sm *ShardMaster) addNextConfig() {

  lastConfig := sm.configs[len(sm.configs) - 1]
  nextConfig := Config{}

  nextConfig.Num = lastConfig.Num + 1

	nextConfig.Shards = make([]int64, sm.NShards)
  copy(nextConfig.Shards[:], lastConfig.Shards[:])

  nextConfig.Groups = make(map[int64][]string)
  for k,v := range lastConfig.Groups {
    nextConfig.Groups[k] = make([]string, len(v))
    copy(nextConfig.Groups[k], v)
  }

  sm.configs = append(sm.configs, nextConfig)
}

func (sm *ShardMaster) waitForStatus(seq int) {
  to := 10 * time.Millisecond
  for !sm.dead {
    decided,_ := sm.px.Status(seq)
    if decided {
      return
    }
    time.Sleep(to)
    if to < 10 * time.Second {
      to *= 2
    }
  }
}

// please don't change this function.
func (sm *ShardMaster) Kill() {
  sm.dead = true
  sm.l.Close()
  sm.px.Kill()
}

//
// servers[] contains the ports of the set of
// servers that will cooperate via Paxos to
// form the fault-tolerant shardmaster service.
// me is the index of the current server in servers[].
//
func StartServer(NShards int, servers []string, me int) *ShardMaster {
  gob.Register(JoinArgs{})
  gob.Register(JoinReply{})
  gob.Register(LeaveArgs{})
  gob.Register(LeaveReply{})
  gob.Register(MoveArgs{})
  gob.Register(MoveReply{})
  gob.Register(QueryArgs{})
  gob.Register(QueryReply{})

  sm := new(ShardMaster)
  sm.me = me

	sm.NShards = NShards

  sm.configs = make([]Config, 1)
  sm.configs[0].Groups = map[int64][]string{}
	sm.configs[0].Shards = make([]int64, sm.NShards)

  rpcs := rpc.NewServer()
  rpcs.Register(sm)

  sm.px = paxos.Make(false, false, servers, me, rpcs)

  sm.mu = sync.Mutex{}

  sm.assignments = make(map[int64]map[int]bool)
  sm.pxMaxSeq = -1

  os.Remove(servers[me])
  l, e := net.Listen("unix", servers[me]);
  if e != nil {
    log.Fatal("listen error: ", e);
  }
  sm.l = l

  // please do not change any of the following code,
  // or do anything to subvert it.

  go func() {
    for sm.dead == false {
      conn, err := sm.l.Accept()
      if err == nil && sm.dead == false {
        if sm.unreliable && (rand.Int63() % 1000) < 100 {
          // discard the request.
          conn.Close()
        } else if sm.unreliable && (rand.Int63() % 1000) < 200 {
          // process the request but force discard of reply.
          c1 := conn.(*net.UnixConn)
          f, _ := c1.File()
          err := syscall.Shutdown(int(f.Fd()), syscall.SHUT_WR)
          if err != nil {
            fmt.Printf("shutdown: %v\n", err)
          }
          go rpcs.ServeConn(conn)
        } else {
          go rpcs.ServeConn(conn)
        }
      } else if err == nil {
        conn.Close()
      }
      if err != nil && sm.dead == false {
        fmt.Printf("ShardMaster(%v) accept: %v\n", me, err.Error())
        sm.Kill()
      }
    }
  }()

  return sm
}
